from typing import Protocol

from torch.nn import Module
from torchinfo import summary


class DescribableModel(Protocol):
    @property
    def model(self) -> Module:
        """Returns the underlying PyTorch model."""
        ...

    @property
    def input_size(self) -> int:
        """Returns the size of the input."""
        ...


def describe(model: DescribableModel) -> None:
    print(summary(model.model, input_size=(1, model.input_size)))



from typing import Protocol, runtime_checkable, Sequence
from helpers.maths import Matrix, Vector


class State(Protocol):
    def __call__(self, t: float) -> Vector:
        """Returns the configuration of the robot arm joints at the given time point."""
        ...

    def dot(self, t: float) -> Vector:
        """Returns the velocities of the robot arm joints at the given time point."""
        ...


class Position(Protocol):
    def __call__(self, q: State) -> Vector:
        """Returns the position of the robot arm at the given configuration."""
        ...


class Jacobian(Protocol):
    def __call__(self, q: State) -> Matrix:
        """Returns the Jacobian matrix of the robot arm at the given configuration."""
        ...


class Gravity(Protocol):
    def __call__(self, q: State) -> Vector:
        """Returns the gravity vector of the robot arm at the given configuration."""
        ...


class ControlSignal(Protocol):
    def __call__(
        self, *, q: State, p: Position, J: Jacobian, g: Gravity, t: float
    ) -> Vector:
        """Returns the control signal for the robot arm at the given time point.

        Args:
            q: The state of the robot arm joints, including the configuration and velocities.
            p: The position of the robot arm.
            J: The Jacobian matrix of the robot arm.
            g: The gravity force acting on the robot arm.
            t: The current time point.

        Returns:
            The control signal for the robot arm for the given time point.

        Note:
            The gravity force `g` would only be available if a realistic model of the robot arm is
            used. This would require the parameters of the robot arm to also be known.
        """
        ...


@runtime_checkable
class StatefulControlSignal(ControlSignal, Protocol):
    def reset(self) -> None:
        """Resets the state of the controller. This will always be called before a simulation
        starts."""
        ...


class DynamicsSolver(Protocol):
    def __call__(
        self,
        dynamics: "RobotDynamics",
        *,
        initial_conditions: Sequence[float] | Vector,
        t_range: tuple[float, float],
        t_evaluation: Vector,
    ) -> Matrix:
        """Solves the robot dynamics over the given time span. In the end, the solver should return the
        states of the system at the specified evaluation time points.

        Args:
            dynamics: The robot dynamics to solve.
            initial_conditions: The initial state of the robot arm.
            t_range: The time span over which to solve the dynamics.
            t_evaluation: The time points at which to evaluate the states of the system.

        Returns:
            The states of the system at the specified evaluation time points. The result is a matrix
            where each row corresponds to a particular state variable and each column corresponds to
            the entire state of the system at a particular time point.
        """
        ...


class DynamicsJacobian(Protocol):
    def __call__(self, t: float, state: Vector) -> Matrix:
        """The Jacobian matrix of the right-hand side of the robot dynamics.

        Args:
            t: The current time point.
            state: The current state of the robot arm.

        Returns:
            The Jacobian matrix of the right-hand side of the robot dynamics. This matrix must be
            $n \\times n$, where $n$ is the number of states in the system. Entry $(i, j)$ of the
            matrix must be the partial derivative of the $i$-th element of the right-hand side with
            respect to the $j$-th state variable.
        """
        ...


class RobotDynamics(Protocol):
    def __call__(self, t: float, state: Vector) -> Vector:
        """The dynamics of the robot arm.

        Args:
            t: The current time point.
            state: The current state of the robot arm.

        Returns:
            The derivative of the state vector.

        Note:
            What the state vector contains and looks like depends on the particular implementation of
            the robot arm dynamics, however, it typically includes the joint angles and velocities in
            some form.

            **Important**: Make sure the following implementations are consistent in how they represent
            the state vector:<br>
            - The `__call__` method of the `RobotDynamics` protocol.<br>
            - The `results_from` method of the `SimulatableRobotArm` protocol.<br>
            - The `initial_conditions` property of the `SimulatableRobotArm` protocol
        """
        ...

    def jacobian(self) -> DynamicsJacobian | None:
        """Returns a function that computes the Jacobian of the right-hand side of the robot dynamics.

        See the [`DynamicsJacobian`](exercise_2.md#helpers.robot.simulate.Jacobian) protocol for more
        details on how it should behave. If the Jacobian is not available, this method should return
        `None` and a numerical approximation will automatically be used instead.
        """
        ...

    def solver(self) -> DynamicsSolver | None:
        """Returns the solver for the robot dynamics. If the default solver should be used, this method
        should simply return `None`.
        """
        ...



from typing import Sequence, Protocol, Iterable
from dataclasses import dataclass, field
from helpers.maths import Vector
from helpers.robot.simulate import (
    ControlSignal,
    State,
    Position,
    Jacobian,
    Gravity,
)

import numpy as np


class Trajectory(Protocol):
    def __call__(self, t: float) -> Vector:
        """Returns the target setpoint for the given time point."""
        ...


class ControlSignalProvider(Protocol):
    def __call__(self, target: Vector, /) -> ControlSignal:
        """Returns a control signal for the given target setpoint."""
        ...


class DistanceMeasure(Protocol):
    def __call__(self, *, q: State, p: Position, t: float, target: Vector) -> float:
        """Returns the distance between the current position (not necessarily in task space) and the
        target setpoint."""
        ...


class ControlSignalWithoutModel(Protocol):
    def __call__(self, *, q: State, p: Position, J: Jacobian, t: float) -> Vector:
        """Returns the control signal for the robot arm at the given time point.

        Args:
            q: The state of the robot arm joints, including the configuration and velocities.
            p: The position of the robot arm.
            J: The Jacobian matrix of the robot arm.
            t: The current time point.

        Returns:
            The control signal for the robot arm for the given time point.

        Note:
            In this case, information about the gravity force acting on the robot arm is not provided.
            This would typically be the case when none of the system parameters are known, except for
            easily measurable physical quantities like the dimensions of the robot arm, the joint angles,
            and the joint velocities.
        """
        ...


class ControlSignalWithTime(Protocol):
    def __call__(self, t: float, /) -> Vector:
        """Returns the control signal for the robot arm at the given time point.

        Args:
            t: The current time point.

        Returns:
            The control signal for the robot arm for the given time point.

        Note:
            This is useful when the control signal is time-dependent and does not require any information
            about the state of the robot arm.
        """
        ...


@dataclass
class SetpointFollower:
    setpoints: Sequence[Vector]
    control_provider: ControlSignalProvider
    distance: DistanceMeasure
    threshold: float
    current_setpoint_index: int = field(init=False, default=0)
    current_control: ControlSignal = field(init=False)

    def __post_init__(self) -> None:
        assert len(self.setpoints) > 0, "At least one setpoint must be provided."

        self.update_control_signal()

    def __call__(
        self, q: State, p: Position, J: Jacobian, g: Gravity, t: float
    ) -> Vector:
        while self.setpoints_remaining() and self.setpoint_reached(q, p, t):
            self.advance_setpoint()

        return self.current_control(q=q, p=p, J=J, g=g, t=t)

    def reset(self) -> None:
        self.current_setpoint_index = 0
        self.update_control_signal()

    def advance_setpoint(self) -> None:
        self.current_setpoint_index += 1
        self.update_control_signal()

    def update_control_signal(self) -> None:
        self.current_control = self.control_provider(self.current_setpoint)

    def setpoints_remaining(self) -> bool:
        return self.current_setpoint_index < (len(self.setpoints) - 1)

    def setpoint_reached(self, q: State, p: Position, t: float) -> bool:
        return (
            self.distance(q=q, p=p, t=t, target=self.current_setpoint) < self.threshold
        )

    @property
    def current_setpoint(self) -> Vector:
        return self.setpoints[self.current_setpoint_index]


@dataclass(frozen=True)
class TrajectoryFollower:
    trajectory: Trajectory
    control_provider: ControlSignalProvider

    def __call__(
        self, q: State, p: Position, J: Jacobian, g: Gravity, t: float
    ) -> Vector:
        return self.control_provider(self.trajectory(t))(q=q, p=p, J=J, g=g, t=t)


def without_model(tau: ControlSignalWithoutModel) -> ControlSignal:
    """Makes a control signal without model information.

    Args:
        tau: The control signal without model.

    Returns:
        The control signal with model.

    Example:
        If your control logic does not require model information, you can use this function to
        create a valid control signal as follows:

        ```python
        # Import the control module

        def signal(q: State, p: Position, J: Jacobian, t: float) -> Vector:
            # Your control logic here
            return ...

        tau = control.without_model(signal)

        # Now you can use `tau` as a regular control signal.
        results = robot.simulate(t, tau)
    """
    return lambda q, p, J, g, t: tau(q=q, p=p, J=J, t=t)


def with_time(tau: ControlSignalWithTime) -> ControlSignal:
    """Makes a time-dependent control signal.

    Args:
        tau: The time-dependent control signal.

    Returns:
        The control signal with model.

    Example:
        If your control logic depends only on the current time point, you can use this function to
        create a valid control signal as follows:

        ```python
        # Import the control module

        def signal(t: float) -> Vector:
            # Your control logic here
            return ...

        tau = control.with_time(signal)

        # Now you can use `tau` as a regular control signal.
        results = robot.simulate(t, tau)
    """
    return lambda q, p, J, g, t: tau(t)


def zero(n: int) -> ControlSignal:
    """Generates a zero control signal.

    Args:
        n: The number of joints of the robot arm.

    Returns:
        The zero control signal that can be directly used as a control signal.
    """

    return lambda q, p, J, g, t: np.zeros(n)


def constant(value: Vector | Iterable[float]) -> ControlSignal:
    """Generates a constant control signal.

    Args:
        value: The constant value of the control signal.

    Returns:
        The constant control signal that can be directly used as a control signal.
    """

    value = np.array(value)

    return lambda q, p, J, g, t: value


def step(*, amplitudes: Sequence[float], t_range: tuple[float, float]) -> ControlSignal:
    """Generates a step signal.

    Args:
        amplitudes: The amplitudes of the step signals for each joint.
        step_time: The time at which the step occurs.

    Returns:
        The step signal that can be directly used as a control signal.
    """

    amplitude = np.array(amplitudes)
    t_0, _ = t_range

    def signal(t_c: float) -> Vector:
        return (amplitude * (t_c >= t_0)).astype(float)

    return lambda q, p, J, g, t: signal(t)


def ramp(*, slopes: Sequence[float], t_range: tuple[float, float]) -> ControlSignal:
    """Generates a ramp signal.

    Args:
        slopes: The slopes of the ramp signals for each joint.
        start_time: The time at which the ramp starts (default is 0).
        end_time: The time at which the ramp ends (default is the end of the time vector).

    Returns:
        The ramp signal that can be directly used as a control signal.
    """

    slope = np.array(slopes)
    t_0, t_f = t_range

    def signal(t_c: float) -> Vector:
        if t_c < t_0:
            return np.zeros_like(slope)
        elif t_c > t_f:
            return slope * (t_f - t_0)
        else:
            return slope * (t_c - t_0)

    return lambda q, p, J, g, t: signal(t)


def chirp(
    *,
    amplitudes: Sequence[float],
    start_frequencies: Sequence[float],
    end_frequencies: Sequence[float],
    fading_factor: float = 2,
    t_range: tuple[float, float],
) -> ControlSignal:
    """Generates a chirp signal.

    Args:
        amplitudes: The amplitudes of the chirp signals for each joint.
        start_frequencies: The start frequencies of the chirp signals for each joint.
        end_frequency: The end frequency of the chirp signals for each joint.
        fading_factor: The fading factor for the chirp signal (default is 2).
        t_range: The time range of the chirp signal.

    Returns:
        The chirp signal that can be directly used as a control signal.
    """

    amplitude = np.array(amplitudes)
    start = np.array(start_frequencies)
    end = np.array(end_frequencies)

    t_0, t_f = t_range
    delta_t = t_f - t_0

    def signal(t_c: float) -> Vector:
        frequencies = start + (end - start) * (t_c - t_0) / delta_t
        return (
            amplitude
            * np.sin(2 * np.pi * frequencies * t_c)
            * (1 - np.exp(-((t_f - t_c) ** 2) * fading_factor))
        )

    return lambda q, p, J, g, t: signal(t)



from typing import Protocol, Sequence
from dataclasses import dataclass


import numpy as np


class Tank(Protocol):
    @property
    def max_inlet_flow(self) -> float:
        """Returns the maximum inflow rate of the tank."""
        ...


@dataclass(frozen=True)
class TankSystemState:
    """This represents the state of the tank system.

    Attributes:
        height_difference: The difference in height between the desired and
            actual liquid levels in the tank.
    """

    height_difference: float

    def __str__(self) -> str:
        return f"Δh = {self.height_difference} m"

    def __lt__(self, other: "TankSystemState") -> bool:
        return self.height_difference < other.height_difference

    @property
    def value(self) -> float:
        return self.height_difference


@dataclass(frozen=True)
class TankSystemAction:
    """This represents an action for the tank system.

    Attributes:
        inflow_rate: The inflow rate of the tank.
    """

    inflow_rate: float

    def __str__(self) -> str:
        return f"u = {self.inflow_rate} m³/s"

    def __lt__(self, other: "TankSystemAction") -> bool:
        return self.inflow_rate < other.inflow_rate

    @property
    def value(self) -> float:
        return self.inflow_rate


class DiscretizeTankMixin:
    def actions(self: Tank, delta_u: float) -> Sequence[TankSystemAction]:
        """Returns a sequence of all possible actions for the tank.

        Args:
            delta_u: The discretization step size for the inflow rate.

        Returns:
            A sequence of all possible actions for the tank.
        """
        u_max = self.max_inlet_flow
        points = int(u_max / delta_u) + 1

        return tuple(
            TankSystemAction(inflow_rate=inflow_rate)
            for inflow_rate in np.linspace(0, u_max, points)
        )



from typing import TypeAlias, Sequence
from dataclasses import dataclass, KW_ONLY

from helpers.maths import Vector, IntVector

from plotly.graph_objects import Scatter3d, Surface, Mesh3d
from scipy.spatial.transform import Rotation

import numpy as np


Trace: TypeAlias = Scatter3d | Surface | Mesh3d


@dataclass(frozen=True)
class Cylinder:
    _: KW_ONLY
    center: tuple[float, float, float]
    normal: tuple[float, float, float]
    radius: float
    height: float
    color: str
    name: str
    group: str | None = None
    text: str | None = None
    angle_points: int = 360
    radial_points: int = 2
    height_points: int = 2

    def traces(self) -> Sequence[Trace]:
        cylinder = self.cylinder_trace()

        if self.text:
            return cylinder, self.text_trace(self.text)
        else:
            return (cylinder,)

    def cylinder_trace(self) -> Trace:
        x, y, z = cylinder_coordinates(
            center=self.center,
            normal=self.normal,
            radius=self.radius,
            height=self.height,
            angle_points=self.angle_points,
            height_points=self.height_points,
        )
        i, j, k = cylinder_triangles(rows=z.shape[0], columns=z.shape[1])

        return Mesh3d(
            x=x.flatten(),
            y=y.flatten(),
            z=z.flatten(),
            i=i,  # triangle indices
            j=j,
            k=k,
            color=self.color,
            name=self.name,
            legendgroup=self.effective_group,
            legendgrouptitle=dict(text=""),
            showlegend=True,
        )

    def text_trace(self, text: str) -> Trace:
        return Scatter3d(
            x=[self.center[0]],
            y=[self.center[1]],
            z=[self.center[2]],
            mode="text",
            text=[text],
            textposition="middle center",
            name=self.name,
            legendgroup=self.effective_group,
            legendgrouptitle=dict(text=""),
            showlegend=False,
        )

    @property
    def effective_group(self) -> str:
        return self.group if self.group is not None else self.name


def cylinder_coordinates(
    *,
    center: tuple[float, float, float],
    normal: tuple[float, float, float],
    radius: float,
    height: float,
    angle_points: int,
    height_points: int,
) -> tuple[Vector, Vector, Vector]:
    x, y, z = cylinder_grid(
        radius=radius,
        height=height,
        angle_points=angle_points,
        height_points=height_points,
    )
    x_rotated, y_rotated, z_rotated = rotate(
        normal, x=x.flatten(), y=y.flatten(), z=z.flatten()
    )
    x, y, z = (
        x_rotated.reshape(x.shape),
        y_rotated.reshape(y.shape),
        z_rotated.reshape(z.shape),
    )

    return x + center[0], y + center[1], z + center[2]


def cylinder_triangles(
    *, rows: int, columns: int
) -> tuple[IntVector, IntVector, IntVector]:
    # Side triangles
    row_indices = np.repeat(np.arange(rows - 1), columns - 1)
    column_indices = np.tile(np.arange(columns - 1), rows - 1)
    base = row_indices * columns + column_indices

    i_1 = base
    j_1 = base + 1
    k_1 = base + columns

    i_2 = base + 1
    j_2 = base + columns + 1
    k_2 = base + columns

    # Bottom cap
    bottom_i = np.zeros(columns - 1, dtype=np.int32)
    bottom_j = np.arange(1, columns)
    bottom_k = np.arange(2, columns + 1)

    # Top cap
    top_center = (rows - 1) * columns
    top_i = np.full(columns - 1, top_center, dtype=np.int32)
    top_j = top_center + np.arange(columns - 1)
    top_k = top_center + np.arange(1, columns)

    i = np.concatenate([i_1, i_2, bottom_i, top_i])
    j = np.concatenate([j_1, j_2, bottom_j, top_j])
    k = np.concatenate([k_1, k_2, bottom_k, top_k])

    return i, j, k


def cylinder_grid(
    *, radius: float, height: float, angle_points: int, height_points: int
) -> tuple[Vector, Vector, Vector]:
    h = np.linspace(-height / 2, height / 2, height_points)
    theta = np.linspace(0, 2 * np.pi, angle_points)
    theta_grid, h_grid = np.meshgrid(theta, h)

    x = radius * np.cos(theta_grid)
    y = radius * np.sin(theta_grid)
    z = h_grid

    return x, y, z


def normalize(vector: tuple[float, float, float]) -> Vector:
    array = np.array(vector)
    return array / np.linalg.norm(array)


def rotate(
    normal: tuple[float, float, float], *, x: Vector, y: Vector, z: Vector
) -> tuple[Vector, Vector, Vector]:
    # Find rotation matrix from [0,0,1] to normal vector
    rotation = Rotation.align_vectors([normalize(normal)], [[0, 0, 1]])[0]

    # Rotate all points
    points = np.stack([x, y, z]).T
    rotated = rotation.apply(points)

    return rotated[:, 0], rotated[:, 1], rotated[:, 2]



from typing import Sequence
from dataclasses import dataclass

from helpers.maths import Vector, Matrix
from helpers.robot.simulate.types import (
    RobotDynamics,
)

from scipy.integrate import solve_ivp

import numpy as np


@dataclass(frozen=True)
class DivergenceChecker:
    divergence_threshold: float = 1e3

    def __call__(self, t: float, state: Vector) -> float:
        return 0.0 if np.linalg.norm(state) > self.divergence_threshold else 1.0

    @property
    def terminal(self) -> bool:
        return True


@dataclass(frozen=True)
class LsodaSolver:
    checker: DivergenceChecker = DivergenceChecker()

    @staticmethod
    def with_divergence_threshold(threshold: float) -> "LsodaSolver":
        """Creates an LSODA solver with the specified divergence threshold.

        Args:
            threshold: The divergence threshold for the solver.

        Returns:
            A new LSODA solver with the specified divergence threshold.
        """
        return LsodaSolver(checker=DivergenceChecker(threshold))

    def __call__(
        self,
        dynamics: RobotDynamics,
        *,
        initial_conditions: Sequence[float] | Vector,
        t_range: tuple[float, float],
        t_evaluation: Vector,
    ) -> Matrix:
        result = solve_ivp(
            fun=dynamics,
            jac=dynamics.jacobian(),
            y0=initial_conditions,
            t_span=t_range,
            t_eval=t_evaluation,
            method="LSODA",
            events=self.checker,
        )

        if result.status == 1:
            raise RuntimeError(
                "The simulation diverged. The system is unstable or the initial conditions are invalid."
            )

        return result.y


class EulerSolver:
    def __call__(
        self,
        dynamics: RobotDynamics,
        *,
        initial_conditions: Sequence[float] | Vector,
        t_range: tuple[float, float],
        t_evaluation: Vector,
    ) -> Matrix:
        states = []
        last_state = np.array(initial_conditions)
        last_t = t_range[0]

        for t in t_evaluation:
            delta_t = t - last_t
            derivative = dynamics(last_t, last_state)
            last_state = last_state + delta_t * derivative

            states.append(last_state)

        return np.array(states).T



from dataclasses import dataclass, field
from enum import Enum

from helpers.maths import BoolMatrix, IntMatrix
from helpers.rl.game import GridGame

from plotly.graph_objects import Figure, Scatter, Heatmap

import numpy as np


class TraceName(Enum):
    UNCOVERED = "Uncovered"
    PLAYER = "Player"
    POSITIVE_REWARDS = "Positive Rewards"
    NEGATIVE_REWARDS = "Negative Rewards"


@dataclass(frozen=True)
class GridGameVisualizer:
    game: GridGame
    trace_indices_by_name: dict[TraceName, int] = field(default_factory=dict)

    def create(self) -> Figure:
        figure = Figure()

        figure, uncovered_index = self._draw_grid_on(figure)
        figure, player_index = self._draw_player_marker(figure)

        figure, positive_rewards_index = self._draw_rewards(
            figure,
            mask=self.game.uncovered,
            condition=self.game.rewards > 0,
            rewards=self.game.rewards,
            color="green",
        )

        figure, negative_rewards_index = self._draw_rewards(
            figure,
            mask=self.game.uncovered,
            condition=self.game.rewards < 0,
            rewards=self.game.rewards,
            color="red",
        )

        self._configure_layout(figure)

        self.trace_indices_by_name.clear()
        self.trace_indices_by_name.update(
            {
                TraceName.UNCOVERED: uncovered_index,
                TraceName.PLAYER: player_index,
                TraceName.POSITIVE_REWARDS: positive_rewards_index,
                TraceName.NEGATIVE_REWARDS: negative_rewards_index,
            }
        )

        return figure

    def update(self, figure: Figure) -> Figure:
        self._update_uncovered_tiles(figure)
        self._update_player_marker(figure)
        self._update_rewards(
            figure,
            self.game.rewards > 0,
            self.trace_indices_by_name[TraceName.POSITIVE_REWARDS],
        )
        self._update_rewards(
            figure,
            self.game.rewards < 0,
            self.trace_indices_by_name[TraceName.NEGATIVE_REWARDS],
        )

        return figure

    def _draw_grid_on(self, figure: Figure) -> tuple[Figure, int]:
        """Draws the grid on the figure and returns the trace index of the trace representing the
        uncovered tiles."""
        traces = trace_count_of(figure)

        self._draw_base_tiles(figure)
        self._draw_uncovered_tiles(figure)
        self._draw_grid_lines(figure)

        return figure, traces + 1

    def _draw_base_tiles(self, figure: Figure) -> Figure:
        figure.add_trace(
            Heatmap(
                z=np.ones_like(self.game.rewards),
                showscale=False,
                colorscale=[[0, "#e0e0e0"], [1, "#e0e0e0"]],  # light gray
                hoverongaps=False,
                hoverinfo="skip",
            )
        )

        return figure

    def _draw_uncovered_tiles(self, figure: Figure) -> Figure:
        uncovered = np.where(self.game.uncovered, 1, np.nan)
        figure.add_trace(
            Heatmap(
                z=uncovered,
                showscale=False,
                colorscale=[[0, "#a0a0a0"], [1, "#a0a0a0"]],  # darker gray
                hoverongaps=False,
                hoverinfo="skip",
            )
        )

        return figure

    def _draw_grid_lines(self, figure: Figure) -> Figure:
        for i in range(self.game.rows + 1):
            figure.add_shape(
                type="line",
                x0=-0.5,
                x1=self.game.columns - 0.5,
                y0=i - 0.5,
                y1=i - 0.5,
                line=dict(color="white", width=2),
            )

        for i in range(self.game.columns + 1):
            figure.add_shape(
                type="line",
                x0=i - 0.5,
                x1=i - 0.5,
                y0=-0.5,
                y1=self.game.rows - 0.5,
                line=dict(color="white", width=2),
            )

        return figure

    def _draw_player_marker(self, figure: Figure) -> tuple[Figure, int]:
        """Draws the player marker on the figure and returns the trace index of the trace."""
        traces = trace_count_of(figure)

        row, column = self.game.current_position
        figure.add_trace(
            Scatter(
                x=[column],
                y=[row],
                mode="markers",
                marker=dict(
                    symbol="circle",
                    size=25,
                    color="rgba(255, 255, 255, 0.75)",
                    line=dict(color="black", width=2),
                ),
                name="Player",
                hoverinfo="skip",
            )
        )

        return figure, traces

    def _draw_rewards(
        self,
        figure: Figure,
        *,
        mask: BoolMatrix,
        condition: BoolMatrix,
        rewards: IntMatrix,
        color: str,
    ) -> tuple[Figure, int]:
        """Draws the rewards on the figure and returns the trace index of the trace."""
        traces = trace_count_of(figure)
        full_mask = mask & condition
        row_idx, col_idx = np.where(full_mask)

        figure.add_trace(
            Scatter(
                x=col_idx,
                y=row_idx,
                text=rewards[full_mask],
                mode="text",
                textfont=dict(color=color, size=20),
                hoverinfo="none",
            )
        )

        return figure, traces

    def _configure_layout(self, figure: Figure) -> None:
        figure.update_layout(
            xaxis=dict(
                showgrid=False,
                range=[-0.5, self.game.columns - 0.5],
                constrain="domain",
                showticklabels=False,
                fixedrange=True,
            ),
            yaxis=dict(
                showgrid=False,
                range=[-0.5, self.game.rows - 0.5],
                scaleanchor="x",
                constrain="domain",
                showticklabels=False,
                fixedrange=True,
            ),
            width=225,
            height=225,
            showlegend=False,
            margin=dict(l=0, r=0, t=0, b=0),
            paper_bgcolor="rgba(0,0,0,0)",
            plot_bgcolor="rgba(0,0,0,0)",
        )

    def _update_uncovered_tiles(self, figure: Figure) -> None:
        uncovered = np.where(self.game.uncovered, 1, np.nan)
        figure.data[self.trace_indices_by_name[TraceName.UNCOVERED]].z = uncovered  # type: ignore

    def _update_player_marker(self, figure: Figure) -> None:
        row, column = self.game.current_position
        figure.data[self.trace_indices_by_name[TraceName.PLAYER]].x = [column]  # type: ignore
        figure.data[self.trace_indices_by_name[TraceName.PLAYER]].y = [row]  # type: ignore

    def _update_rewards(
        self, figure: Figure, condition: BoolMatrix, trace_index: int
    ) -> None:
        full_mask = self.game.uncovered & condition
        row_idx, col_idx = np.where(full_mask)

        figure.data[trace_index].x = col_idx  # type: ignore
        figure.data[trace_index].y = row_idx  # type: ignore
        figure.data[trace_index].text = self.game.rewards[full_mask]  # type: ignore


def trace_count_of(figure: Figure) -> int:
    return len(figure.data) if figure.data else 0  # type: ignore


def annotation_count_of(figure: Figure) -> int:
    return len(figure.layout.annotations) if figure.layout.annotations else 0  # type: ignore


def trace_of(figure: Figure, index: int) -> dict:
    return figure.data[index]  # type: ignore



from typing import Protocol
from dataclasses import dataclass

import os
import pandas as pd


class DataMapper(Protocol):
    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        """Performs some transformation on the data."""
        ...


@dataclass(frozen=True)
class DataLoader:
    data_by_filename: dict[str, pd.DataFrame]

    @staticmethod
    def for_directory(directory: str) -> "DataLoader":
        return DataLoader(
            data_by_filename={
                filename: pd.read_csv(f"{directory}/{filename}")
                for filename in os.listdir(directory)
            }
        )

    def statistics(self) -> pd.DataFrame:
        return self.combine().describe()

    def one(self) -> pd.DataFrame:
        return next(iter(self.data_by_filename.values()))

    def map(self, mapper: DataMapper) -> "DataLoader":
        return DataLoader(
            data_by_filename={
                filename: mapper(data)
                for filename, data in self.data_by_filename.items()
            }
        )

    def combine(self) -> pd.DataFrame:
        return pd.concat(self.data_by_filename.values())



from typing import Final, Sequence, Self
from dataclasses import dataclass, field, KW_ONLY, replace

from helpers.maths import Point, Matrix, Vector
from helpers.robot.arm import RobotArmMixin
from helpers.robot.link import Link
from helpers.robot.joint import Joint
from helpers.robot.jacobian import RobotJacobianMixin
from helpers.robot.planar.animation import (
    PlanarRobotAnimationMixin,
)
from helpers.robot.planar.dynamics import RigidBodyDynamics
from helpers.robot.predict import RobotDynamicsProvider, RobotArmPredictionMixin
from helpers.robot.simulate import (
    SimulationResults,
    RobotArmSimulationMixin,
    ControlSignal,
    AnimatedRobotArm,
    RobotDynamics,
)

import numpy as np


@dataclass
class PlanarRobotArm(
    PlanarRobotAnimationMixin,
    RobotArmMixin,
    RobotArmSimulationMixin,
    RobotArmPredictionMixin,
    RobotJacobianMixin,
    AnimatedRobotArm,
):
    _: KW_ONLY
    name: Final[str] = "Helpful Roboto :)"

    link_1: Link
    link_2: Link
    joint_1: Joint = Joint(theta=0.0)
    joint_2: Joint = Joint(theta=0.0)

    rotations: int = field(default=0, init=False)
    dynamics: RobotDynamicsProvider = field(default=RigidBodyDynamics)

    def using_dynamics_from(self, dynamics: RobotDynamicsProvider) -> Self:
        """Returns a new robot arm that uses the dynamics model from the given provider.

        Args:
            dynamics: The provider of the dynamics model for the robot arm.

        Returns:
            A new robot arm that uses the dynamics model from the given provider.

        Note:
            See the [`RobotDynamicsProvider`](exercise_2.md#helpers.robot.RobotDynamicsProvider)
            protocol for more information on how exactly the provider should work.
        """
        return replace(self, dynamics=dynamics)

    def joint_positions(self) -> list[Point]:
        return self.forward_kinematics((self.theta_1, self.theta_2))

    def forward_kinematics(self, q: Sequence[float] | Vector) -> list[Point]:
        assert len(q) == 2, f"Expected 2 joint angles. Got: {len(q)}"

        theta_1, theta_2 = q

        x_0, y_0 = 0, 0
        x_1, y_1 = (
            x_0 + self.l_1 * np.cos(theta_1),
            y_0 + self.l_1 * np.sin(theta_1),
        )
        x, y = (
            x_1 + self.l_2 * np.cos(theta_1 + theta_2),
            y_1 + self.l_2 * np.sin(theta_1 + theta_2),
        )

        return [Point(x_0, y_0), Point(x_1, y_1), Point(x, y)]

    def dynamics_for(self, tau: ControlSignal) -> RobotDynamics:
        return self.dynamics.create_for(self, torque=tau)

    def results_from(
        self,
        *,
        t_range: tuple[float, float],
        recorded_points: int,
        t: Vector,
        tau: list[Vector],
        solution: Matrix,
    ) -> "SimulationResults[PlanarRobotArm]":
        return self.dynamics.results_for(
            self,
            t_range=t_range,
            recorded_points=recorded_points,
            t=t,
            tau=tau,
            solution=solution,
        )

    def initial_conditions(
        self, initial_velocities: Sequence[float] | None = None
    ) -> Vector:
        initial_velocities = initial_velocities or [0, 0]

        assert len(initial_velocities) == 2, (
            f"The robot has 2 joints, so 2 initial velocities must be provided. Got: {len(initial_velocities)}"
        )

        q = np.array([self.theta_1, self.theta_2])
        q_dot = np.array(initial_velocities)

        return np.concatenate([q, q_dot])

    @property
    def l_1(self) -> float:
        return self.link_1.length

    @property
    def l_2(self) -> float:
        return self.link_2.length

    @property
    def theta_1(self) -> float:
        return self.joint_1.theta

    @theta_1.setter
    def theta_1(self, value: float) -> None:
        self.joint_1 = self.joint_1.with_angle(value)

    @property
    def theta_2(self) -> float:
        return self.joint_2.theta

    @theta_2.setter
    def theta_2(self, value: float) -> None:
        self.joint_2 = self.joint_2.with_angle(value)



from .tank import Tank as Tank
from .policy import Policy as Policy
from .control import Control as Control
from .discretize import (
    TankSystemState as TankSystemState,
    TankSystemAction as TankSystemAction,
)
from .animation import display_setpoint as display_setpoint
from .visualize import (
    visualize_discretization as visualize_discretization,
    visualize_reward_function as visualize_reward_function,
    visualize_policy as visualize_policy,
    visualize_training_results as visualize_training_results,
    theoretical_epsilon_greedy_probabilities as theoretical_epsilon_greedy_probabilities,
)
from .learn import (
    LearningController as LearningController,
    TrainingResult as TrainingResult,
    SetpointRecorder as SetpointRecorder,
)



from typing import Protocol, TypeVar, Generic, Self, Sequence
from dataclasses import dataclass, KW_ONLY  #

from helpers.maths import Matrix, Vector
from helpers.robot.animation import JointAngles
from helpers.robot.simulate.types import (
    ControlSignal,
    RobotDynamics,
    State,
    Position,
    Jacobian,
    Gravity,
)

import os
import numpy as np
import pandas as pd

RobotT = TypeVar("RobotT", bound="SimulatableRobotArm", covariant=True)
InvariantRobotT = TypeVar("InvariantRobotT", bound="SimulatableRobotArm")


@dataclass(frozen=True)
class SimulationResultsExporter:
    results: list["SimulationResults"]

    def to(self, directory: str) -> None:
        """Exports the simulation results to a CSV file.

        Args:
            directory: The directory to save the simulation results to.

        Example:
            ```python
            results_1 = ... # Some simulation results
            results_2 = ... # Some other simulation results

            SimulationResults.export(results_1, results_2).to("path/to/directory")
            # The simulation results will be saved under the specified directory.
            ```
        """
        os.makedirs(directory, exist_ok=True)

        for i, results in enumerate(self.results):
            results.save_to(f"{directory}/results_{i + 1}.csv")


@dataclass(frozen=True)
class SimulationResults(Generic[RobotT]):
    """The results of a robot arm simulation.

    Attributes:
        t: The time points of the simulation.
        tau: The torques applied to the robot arm. Each vector corresponds to a joint.
        q: The joint angles of the robot arm. Each vector corresponds to a joint.
        q_dot: The joint velocities of the robot arm. Each vector corresponds to a joint.
        robot: The robot arm for which the simulation results were generated.

    Example:
        ```python
        results = robot.simulate(t, tau)

        # You can animate the results of the simulation like this:
        results.animate()

        # You can save the results to a file like this:
        results.save_to("path/to/file.csv")
        ```
    """

    _: KW_ONLY
    t: Vector
    tau: list[Vector]
    q: list[Vector]
    q_dot: list[Vector]

    robot: RobotT

    _recorded_points: int
    _t_range: tuple[float, float]
    _final_state: Vector

    @staticmethod
    def of(
        robot: InvariantRobotT,
        *,
        t_range: tuple[float, float],
        recorded_points: int,
        t: Vector,
        tau: list[Vector],
        q: list[Vector],
        q_dot: list[Vector],
        final_state: Vector,
    ) -> "SimulationResults[InvariantRobotT]":
        return SimulationResults(
            t=t,
            tau=tau,
            q=q,
            q_dot=q_dot,
            robot=robot,
            _t_range=t_range,
            _recorded_points=recorded_points,
            _final_state=final_state,
        )

    @staticmethod
    def export(*results: "SimulationResults") -> SimulationResultsExporter:
        """Creates an exporter for the simulation results.

        Args:
            results: The simulation results to be exported.

        Returns:
            The exporter for the simulation results.

        Note:
            See [`SimulationResultsExporter`](exercise_2.md#helpers.robot.simulate.SimulationResultsExporter) for
            more details on how to export the simulation results.
        """
        return SimulationResultsExporter(list(results))

    def animate(
        self: "SimulationResults[AnimatedRobotArm]",
        *,
        take: int = 1,
    ) -> "SimulationResults[AnimatedRobotArm]":
        """Animates the results of the robot arm simulation.

        Args:
            take: The stride to take when animating the simulation results. A stride of `1` will animate every
                recorded point, a stride of `2` will animate every other recorded point, and so on.

        Note:
            This will only work if the robot for which the simulation results were generated is an
            [`AnimatedRobotArm`](exercise_2.md#helpers.robot.simulate.AnimatedRobotArm).
        """

        self.robot.animate(JointAngles.combining(self.q[0], self.q[1])[::take])
        return self

    def save_to(self, file: str) -> None:
        """Saves the simulation results to a file as a CSV.

        Args:
            file: The file path to save the simulation results to.
        """
        pd.DataFrame(
            {
                "Time (s)": self.t,
                **{f"Joint {i + 1} Torque (Nm)": tau for i, tau in enumerate(self.tau)},
                **{f"Joint {i + 1} Angle (rad)": q for i, q in enumerate(self.q)},
                **{
                    f"Joint {i + 1} Velocity (rad/s)": q_dot
                    for i, q_dot in enumerate(self.q_dot)
                },
            }
        ).to_csv(file, index=False, header=True)

    def continue_with(
        self,
        tau: ControlSignal,
        *,
        t_range: tuple[float, float] | None = None,
        recorded_points: int | None = None,
    ) -> "SimulationResults[RobotT]":
        """Continues the simulation with the given control signal.

        Args:
            tau: The control signal to continue the simulation with.
            t: The time span to evaluate the simulation over.

        Returns:
            The simulation results of the continued simulation with the given control signal.

        Note:
            The simulation will continue from the final time point of the current simulation results. If a
            new time span is provided, the simulation will continue over that time span. If no time span is
            provided, the simulation will continue for the same amount of time as the span of the original
            simulation results.

            It does not matter if the specified time span starts at 0 or at the final time point of the
            current simulation results. The simulation will always continue from the final time point of the
            current simulation results.
        """
        recorded_points = recorded_points or self._recorded_points
        t_0, t_f = t_range or self._t_range

        t_span = t_f - t_0
        delta_t = t_span / (recorded_points - 1)
        start_time = self._final_time + delta_t
        start_tau = self._final_tau

        def combined_tau(
            q: State, p: Position, J: Jacobian, g: Gravity, t: float
        ) -> Vector:
            return start_tau if t < start_time else tau(q=q, p=p, J=J, g=g, t=t)

        return self._extend(
            self.robot.simulate(
                combined_tau,
                t_range=(self._final_time, start_time + t_span),
                recorded_points=recorded_points + 1,
                initial_conditions=self._final_state,
                skip_first=True,
            )
        )

    def _extend(
        self, other: "SimulationResults[RobotT]"
    ) -> "SimulationResults[RobotT]":
        """Extends the simulation results with another set of simulation results."""
        return SimulationResults.of(
            self.robot,
            t_range=self._t_range,
            recorded_points=self._recorded_points,
            t=np.concat([self.t, other.t]),
            tau=[
                np.concat([self_tau, other_tau])
                for self_tau, other_tau in zip(self.tau, other.tau)
            ],
            q=[
                np.concat([self_q, other_q]) for self_q, other_q in zip(self.q, other.q)
            ],
            q_dot=[
                np.concat([self_q_dot, other_q_dot])
                for self_q_dot, other_q_dot in zip(self.q_dot, other.q_dot)
            ],
            final_state=other._final_state,
        )

    @property
    def _final_time(self) -> float:
        """Returns the final time of the simulation."""
        return self.t[-1]

    @property
    def _final_tau(self) -> Vector:
        """Returns the final torques of the simulation."""
        return np.array([tau[-1] for tau in self.tau])


class SimulatableRobotArm(Protocol):
    def simulate(
        self,
        tau: ControlSignal,
        *,
        t_range: tuple[float, float],
        recorded_points: int | None = None,
        initial_velocities: Sequence[float] | None = None,
        initial_conditions: Sequence[float] | Vector | None = None,
        skip_first: bool = False,
    ) -> SimulationResults[Self]:
        """Simulates the robot arm dynamics."""
        ...

    def dynamics_for(self, tau: ControlSignal) -> RobotDynamics:
        """Returns the dynamics of the robot arm for the given input torques."""
        ...

    def results_from(
        self,
        *,
        t_range: tuple[float, float],
        recorded_points: int,
        t: Vector,
        tau: list[Vector],
        solution: Matrix,
    ) -> SimulationResults[Self]:
        """Returns the simulation results of the robot arm."""
        ...

    def initial_conditions(self, initial_velocities: Sequence[float] | None) -> Vector:
        """The initial conditions of the robot arm."""
        ...


class AnimatedRobotArm(SimulatableRobotArm, Protocol):
    def animate(self, joint_angles: list[JointAngles]) -> None:
        """Animates the robot arm moving through the given joint angles. See the
        [`animate`](exercise_2.md#helpers.robot.RobotAnimator.animate) function for more details."""
        ...



from typing import Protocol, Sequence

from helpers.ui import basic_animation_configuration
from helpers.tank.state import TankStates
from helpers.tank.geometry import Cylinder, Trace

from plotly.graph_objects import Figure, Frame, Scatter
from plotly.subplots import make_subplots
from concurrent.futures import ProcessPoolExecutor, Future
from multiprocessing import cpu_count
from tqdm import tqdm


class OnFrame(Protocol):
    def __call__(self, figure: Figure, *, index: int, total: int) -> Figure:
        """This function is called after generating each frame of the animation.

        Args:
            figure: The figure representing the current frame.
            index: The index of the current frame.
            total: The total number of frames.

        Note:
            You can assume the figure is fully prepared and ready to be displayed.
        """
        ...


def do_nothing(figure: Figure, *, index: int, total: int) -> Figure:
    """Does nothing."""
    return figure


class Tank(Protocol):
    def draw(self, *, show: bool = True) -> Figure:
        """Draws the tank in its current state."""
        ...

    def draw_state_diagrams(
        self, states: TankStates, *, show: bool = True
    ) -> tuple[Figure, Figure]:
        """Draws the state diagrams of the tank system."""
        ...

    def liquid_height_is(self, height: float) -> "Tank":
        """Returns a new tank with the liquid height set to the given value."""
        ...

    @property
    def name(self) -> str:
        """The name of the tank."""
        ...

    @property
    def height(self) -> float:
        """The height of the tank."""
        ...

    @property
    def diameter(self) -> float:
        """The diameter of the tank."""
        ...

    @property
    def liquid_height(self) -> float:
        """The height of the liquid in the tank."""
        ...

    def _draw_height_diagram(self, states: TankStates) -> Figure:
        """Draws the height diagram of the tank system."""
        ...

    def _draw_flow_diagram(self, states: TankStates) -> Figure:
        """Draws the flow diagram of the tank system."""
        ...

    def _draw_tank_surface(self) -> Sequence[Trace]:
        """Draws the surface of the tank."""
        ...

    def _draw_liquid_surface(self) -> Sequence[Trace]:
        """Draws the surface of the liquid in the tank."""
        ...

    def _draw_inlet(self) -> Sequence[Trace]:
        """Draws the inlet of the tank."""
        ...

    def _draw_outlet(self) -> Sequence[Trace]:
        """Draws the outlet of the tank."""
        ...

    def _configure_layout_of(self, figure: Figure) -> None:
        """Configures the layout of the figure."""
        ...


class VisualizeTankMixin:
    TANK_COLOR = "rgba(0, 0, 0, 0.1)"
    LIQUID_COLOR = "rgba(0, 0, 255, 0.3)"
    INLET_COLOR = "rgba(0, 255, 0, 0.25)"
    OUTLET_COLOR = "rgba(255, 0, 0, 0.25)"

    def animate(
        self: Tank, states: TankStates, *, on_frame: OnFrame = do_nothing
    ) -> None:
        """Animates the tank system over time.

        Args:
            states: The states of the tank system over time.
        """

        frames = animation_frames_for(self, states, on_frame=on_frame)

        def with_progress(figure: Figure, index: int) -> Frame:
            return frame_with_progress_from(
                figure, time=states.time[index - 1], total=states.time[-1]
            )

        prepared_frames = [
            with_progress(frame, index=index)
            for index, frame in enumerate(
                tqdm(frames, desc="Preparing frames", unit=" frame"), start=1
            )
        ]

        figure = animation_figure(frames=prepared_frames)

        figure.show()

    def draw(self: Tank, *, show: bool = True) -> Figure:
        """Draws the tank in its current state.

        Args:
            show: Whether to display the visualization immediately.

        Returns:
            A Plotly figure representing the tank.
        """
        figure = Figure()

        figure.add_traces(self._draw_tank_surface())
        figure.add_traces(self._draw_liquid_surface())
        figure.add_traces(self._draw_inlet())
        figure.add_traces(self._draw_outlet())

        self._configure_layout_of(figure)

        if show:
            figure.show()

        return figure

    def draw_state_diagrams(
        self: Tank, states: TankStates, *, show: bool = True
    ) -> tuple[Figure, Figure]:
        """Draws the state diagrams of the tank system.

        Args:
            states: The states of the tank system.
            show: Whether to display the visualization immediately.

        Returns:
            A tuple containing the height and flow diagrams.
        """
        height_diagram = self._draw_height_diagram(states)
        flow_diagram = self._draw_flow_diagram(states)

        if show:
            height_diagram.show()
            flow_diagram.show()

        return height_diagram, flow_diagram

    def _draw_height_diagram(self: Tank, states: TankStates) -> Figure:
        t_0, t_f = states.time_range
        figure = Figure()

        figure.add_trace(
            Scatter(
                x=states.time,
                y=states.liquid_height,
                mode="lines",
                name="Liquid Height",
                line=dict(color="blue"),
            )
        )

        figure.add_trace(
            Scatter(
                x=[t_0, t_f],
                y=[self.height, self.height],
                mode="lines",
                name="Tank Height",
                line=dict(color="black", dash="dash"),
            )
        )

        figure.update_layout(title="Liquid Height Over Time")

        figure.update_xaxes(title_text="Time (s)", range=states.time_range)
        figure.update_yaxes(title_text="Height (m)", range=[0, self.height * 1.25])

        return figure

    def _draw_flow_diagram(self: Tank, states: TankStates) -> Figure:
        figure = Figure()

        figure.add_trace(
            Scatter(
                x=states.time,
                y=states.inlet_flow_rate,
                mode="lines",
                name="Inlet Flow Rate",
                line=dict(color="green"),
            )
        )

        figure.add_trace(
            Scatter(
                x=states.time,
                y=states.outlet_flow_rate,
                mode="lines",
                name="Outlet Flow Rate",
                line=dict(color="red"),
            )
        )

        figure.update_xaxes(title_text="Time (s)", range=states.time_range)
        figure.update_yaxes(
            title_text="Flow Rate (m³/s)",
            range=[
                0,
                max(states.inlet_flow_rate.max(), states.outlet_flow_rate.max()) * 1.1,
            ],
        )

        figure.update_layout(title="Flow Rates Over Time")

        return figure

    def _draw_tank_surface(self: Tank) -> Sequence[Trace]:
        return Cylinder(
            center=(0, 0, self.height / 2),
            normal=(0, 0, 1),
            radius=self.diameter / 2,
            height=self.height,
            color=VisualizeTankMixin.TANK_COLOR,
            name="Tank",
        ).traces()

    def _draw_liquid_surface(self: Tank) -> Sequence[Trace]:
        return Cylinder(
            center=(0, 0, self.liquid_height / 2),
            normal=(0, 0, 1),
            radius=self.diameter / 2,
            height=self.liquid_height,
            color=VisualizeTankMixin.LIQUID_COLOR,
            name="Liquid",
        ).traces()

    def _draw_inlet(self: Tank) -> Sequence[Trace]:
        return Cylinder(
            center=(0, 0, self.height),
            normal=(0, 0, 1),
            radius=self.diameter / 8,
            height=self.height / 10,
            color=VisualizeTankMixin.INLET_COLOR,
            name="Inlet",
            text="Inlet",
        ).traces()

    def _draw_outlet(self: Tank) -> Sequence[Trace]:
        return Cylinder(
            center=(0, 0, 0),
            normal=(0, 0, 1),
            radius=self.diameter / 8,
            height=self.height / 10,
            color=VisualizeTankMixin.OUTLET_COLOR,
            name="Outlet",
            text="Outlet",
        ).traces()

    def _configure_layout_of(
        self: Tank, figure: Figure, *, padding: float = 0.25
    ) -> None:
        def pad(range_: tuple[float, float]) -> tuple[float, float]:
            return range_[0] - padding, range_[1] + padding

        figure.update_layout(
            title=f"{self.name} the Tank",
            scene=dict(
                xaxis=dict(
                    title="X", range=pad((-self.diameter / 2, self.diameter / 2))
                ),
                yaxis=dict(
                    title="Y", range=pad((-self.diameter / 2, self.diameter / 2))
                ),
                zaxis=dict(title="Z", range=pad((0, self.height))),
                aspectmode="cube",
            ),
        )


def display_setpoint(*, times: Sequence[float], setpoints: Sequence[float]) -> OnFrame:
    """Creates an on-frame callback that displays the setpoint on the tank diagram.

    Args:
        setpoints: The setpoints to display.
    """

    def on_frame(figure: Figure, *, index: int, total: int) -> Figure:
        time_index = int(index * len(times) / (total - 1))

        figure.add_trace(
            Scatter(
                x=times[: time_index + 1],
                y=setpoints[: time_index + 1],
                mode="lines",
                name="Setpoint",
                line=dict(color="red", dash="dash"),
            ),
            row=1,
            col=2,
        )

        return figure

    return on_frame


def animation_frames_for(
    tank: Tank,
    states: TankStates,
    *,
    on_frame: OnFrame,
    min_core_count: int = 4,
    min_state_count: int = 500,
) -> list[Figure]:
    if (cores := cpu_count()) < min_core_count or states.points < min_state_count:
        return [
            on_frame(
                create_frame_for(
                    tank, states.until(i + 1), states.time_range, states.flow_rate_range
                ),
                index=i,
                total=states.points,
            )
            for i in tqdm(
                range(states.points),
                total=states.points,
                desc="Creating animation frames",
                unit=" frame",
            )
        ]
    else:
        with ProcessPoolExecutor(max_workers=cores - 2) as executor:
            futures: list[Future] = []
            frames: list[Figure] = []
            time_range = states.time_range
            flow_rate_range = states.flow_rate_range

            for step in tqdm(
                range(states.points),
                desc="Splitting animation tasks",
                unit=" task",
                total=states.points,
            ):
                future = executor.submit(
                    create_frame_for,
                    tank,
                    states.until(step + 1),
                    time_range,
                    flow_rate_range,
                )
                futures.append(future)

            for i, future in enumerate(
                tqdm(
                    futures,
                    desc="Creating animation frames",
                    unit=" frame",
                    total=states.points,
                )
            ):
                frames.append(on_frame(future.result(), index=i, total=states.points))

            return frames


def create_frame_for(
    tank: Tank,
    states: TankStates,
    time_range: tuple[float, float],
    flow_rate_range: tuple[float, float],
) -> Figure:
    tank_figure = tank.liquid_height_is(states.liquid_height[-1]).draw(show=False)
    height_diagram, flow_diagram = tank.draw_state_diagrams(states, show=False)

    return combine_figures(
        f"{tank.name} the Tank",
        tank_figure=tank_figure,
        height_diagram=height_diagram,
        flow_diagram=flow_diagram,
        time_range=time_range,
        flow_rate_range=flow_rate_range,
    )


def combine_figures(
    title: str,
    *,
    tank_figure: Figure,
    height_diagram: Figure,
    flow_diagram: Figure,
    time_range: tuple[float, float],
    flow_rate_range: tuple[float, float],
) -> Figure:
    def pad(range_: tuple[float, float]) -> tuple[float, float]:
        return range_[0] - 0.1, range_[1] + 0.1

    figure = make_subplots(
        rows=2,
        cols=2,
        column_widths=[0.5, 0.5],
        subplot_titles=["Tank", "Liquid Height", "Flow Rates"],
        specs=[
            [{"type": "scene", "rowspan": 2}, {"type": "xy"}],
            [None, {"type": "xy"}],
        ],
    )

    figure.add_traces(tank_figure.data, rows=1, cols=1)
    figure.add_traces(height_diagram.data, rows=1, cols=2)
    figure.add_traces(flow_diagram.data, rows=2, cols=2)

    figure.update_layout(title=title)
    figure.update_scenes(
        tank_figure.layout.scene,  # type: ignore
        row=1,
        col=1,
    )
    figure.update_xaxes(range=time_range, row=1, col=2)
    figure.update_xaxes(title="Time (s)", range=time_range, row=2, col=2)
    figure.update_yaxes(
        title="Height (m)",
        range=height_diagram.layout.yaxis.range,  # type: ignore
        row=1,
        col=2,
    )
    figure.update_yaxes(
        title="Flow Rate (m³/s)", range=pad(flow_rate_range), row=2, col=2
    )

    return figure


def animation_figure(
    frames: list[Frame],
) -> Figure:
    figure = Figure(data=frames[0].data, layout=frames[0].layout, frames=frames)
    figure.update_layout(
        updatemenus=[basic_animation_configuration(redraw=True, duration=1)],
    )
    return figure


def frame_with_progress_from(figure: Figure, *, time: float, total: float) -> Frame:
    return Frame(
        data=with_progress_information(figure, time=time, total=total).data,
        layout=figure.layout,
    )


def with_progress_information(figure: Figure, *, time: float, total: float) -> Figure:
    title = figure.layout.title.text  # type: ignore

    figure.update_layout(
        title=f"{title} (at t={time:.2f} s / {total:.2f} s)",
    )
    return figure



from typing import NamedTuple, Final, Sequence, Protocol
from dataclasses import dataclass, KW_ONLY

from helpers.maths import IntMatrix, BoolMatrix

import numpy as np


class GridGameSolver(Protocol):
    def solve(self, game: "GridGame") -> "Solution":
        """Find the optimal solution to the given grid game."""
        ...


class Position(NamedTuple):
    row: int
    column: int

    @staticmethod
    def center_of(grid: IntMatrix) -> "Position":
        """Returns the center position of a grid.

        Args:
            grid: The grid to find the center of.

        Returns:
            The center position of the grid.
        """
        rows, columns = grid.shape
        return Position(rows // 2, columns // 2)

    def within_bounds_of(self, grid: IntMatrix) -> bool:
        """Checks if the position is within the bounds of a grid.

        Args:
            grid: The grid to check against.

        Returns:
            True if the position is within the bounds of the grid, False otherwise.
        """
        rows, columns = grid.shape
        return 0 <= self.row < rows and 0 <= self.column < columns


class Move(NamedTuple):
    row: int
    column: int

    def apply(self, position: Position) -> Position:
        """Applies the move to a position.

        Args:
            position: The position to apply the move to.

        Returns:
            The new position after applying the move.
        """
        return Position(position.row + self.row, position.column + self.column)

    def __neg__(self) -> "Move":
        return Move(-self.row, -self.column)


class Moves:
    UP = Move(1, 0)
    DOWN = Move(-1, 0)
    LEFT = Move(0, -1)
    RIGHT = Move(0, 1)

    @staticmethod
    def all() -> Sequence[Move]:
        return (Moves.UP, Moves.DOWN, Moves.LEFT, Moves.RIGHT)


@dataclass(frozen=True)
class Solution:
    """A solution to a grid game.

    Attributes:
        moves: The moves made in the solution.
        reward: The total reward accumulated in the solution.
    """

    moves: Sequence[Move]
    reward: int


# TODO: Test and refactor this class
class GreedySolver:
    def solve(self, game: "GridGame") -> "Solution":
        all_moves = Moves.all()
        best_moves: list[Move] = []
        visited: set[Position] = set()
        total_reward = 0

        while True:
            next_moves: list[tuple[Move, int]] = []
            visited.add(game.current_position)

            # Find all possible moves and their rewards
            for move in all_moves:
                reward = game.move(move)
                next_moves.append((move, reward))
                game.undo()

            # If all moves give negative rewards, stop
            if max(r for _, r in next_moves) < 0:
                break

            # Choose the move with highest reward
            best_move, _ = max(next_moves, key=lambda x: x[1])

            # Make the move
            best_moves.append(best_move)
            total_reward += game.move(best_move)

            # If we're stuck, stop
            if game.current_position in visited:
                break

        return Solution(moves=best_moves, reward=total_reward)


@dataclass(frozen=True)
class StateChange:
    """A change in the state of a grid game.

    Attributes:
        move: The move that was made.
        reward: The reward for making the move.
        position: The new position after making the move.
        moves: The total number of moves made so far.
        accumulated_reward: The total reward accumulated so far.
    """

    start_position: Position
    reward_change: int
    was_covered: bool


class TransitionFunction(Protocol):
    def __call__(self, position: Position, move: Move) -> Position:
        """Calculates the next position after making a move."""
        ...


class RewardFunction(Protocol):
    def __call__(
        self,
        position: Position,
        move: Move,
        transition: TransitionFunction,
        *,
        rewards: IntMatrix,
        visited: BoolMatrix,
    ) -> int:
        """Calculates the reward for a move in the grid game."""
        ...


class EndGame(Exception):
    """This exception is raised when the game should end."""

    ...


class Policy(Protocol):
    def __call__(self, position: Position) -> Move:
        """Returns the move to make from a given position.

        Raises:
            EndGame: If the game should end.
        """
        ...

    def reset(self) -> None:
        """Resets the policy to its initial state.

        This should be called whenever the game is reset."""
        ...


class DeterministicTransition:
    def __call__(self, position: Position, move: Move) -> Position:
        return move.apply(position)


@dataclass
class CachedTransition:
    transition: TransitionFunction
    caching: bool = False
    last: Position | None = None

    def __call__(self, position: Position, move: Move) -> Position:
        if self.caching and self.last is not None:
            return self.last

        self.last = self.transition(position, move)
        return self.last

    def start(self) -> None:
        self.caching = True

    def stop(self) -> None:
        self.caching = False
        self.last = None


@dataclass(frozen=True)
class BoundaryChecked:
    grid: IntMatrix
    transition: TransitionFunction

    def __call__(self, position: Position, move: Move) -> Position:
        new_position = self.transition(position, move)
        return new_position if new_position.within_bounds_of(self.grid) else position


# TODO: Test this class
@dataclass
class GridGame:
    _: KW_ONLY
    safe_transition: Final[BoundaryChecked]
    cached_transition: Final[CachedTransition]
    original_transition: Final[TransitionFunction]
    reward_function: Final[RewardFunction]
    rewards: Final[IntMatrix]
    uncovered: Final[BoolMatrix]
    start_position: Final[Position]
    solver: Final[GridGameSolver]
    policy: Final[Policy | None] = None

    current_position: Position
    moves: int = 0
    accumulated_reward: int = 0
    last_change: StateChange | None = None
    solution: Solution | None = None

    @staticmethod
    def create(
        *,
        rows: int,
        columns: int,
        reward_function: RewardFunction,
        transition_function: TransitionFunction = DeterministicTransition(),
        reward_range: tuple[int, int] = (-5, 5),
        start_position: Position | None = None,
        solver: GridGameSolver = GreedySolver(),
        policy: Policy | None = None,
    ) -> "GridGame":
        """Creates a grid game with random rewards.

        Args:
            rows: The number of rows in the grid.
            columns: The number of columns in the grid.
            reward_function: The function to calculate rewards.
            transition_function: The function to calculate transitions (default is deterministic).
            reward_range: The range of rewards that can be generated (default is (-5, 5)).
            start_position: The starting position of the player in the grid (default is the center).
            solver: The solver to use to find the optimal solution (default is GreedySolver).
            policy: The policy to use for making moves (default is None).

        Returns:
            A grid game with random rewards.
        """
        assert rows > 0, "The number of rows must be positive."
        assert columns > 0, "The number of columns must be positive."
        assert reward_range[0] < reward_range[1], (
            "The lower bound of the reward range must be less than the upper bound."
        )

        rewards = random_rewards(rows=rows, columns=columns, reward_range=reward_range)
        start_position = start_position or Position.center_of(rewards)
        rewards[start_position] = 0

        uncovered = mask_for(rewards, start_position)
        cached_transition = CachedTransition(transition_function)
        game = GridGame(
            safe_transition=BoundaryChecked(rewards, cached_transition),
            cached_transition=cached_transition,
            original_transition=transition_function,
            reward_function=reward_function,
            rewards=rewards,
            uncovered=uncovered,
            start_position=start_position,
            solver=solver,
            policy=policy,
            current_position=start_position,
        )

        solution = solver.solve(game)

        return game.reset().with_solution(solution)

    def move(self, move: Move | None = None) -> int:
        """Moves the player in the grid game.

        Args:
            move: The move to make or None to use the policy.

        Returns:
            The reward for making the move.
        """
        if move is None:
            assert self.policy is not None, (
                "Either a move or a policy must be provided."
            )
            move = self.policy(self.current_position)

        self.cached_transition.start()

        reward = self.reward_function(
            position=self.current_position,
            move=move,
            transition=self.cached_transition,
            rewards=self.rewards,
            visited=self.uncovered,
        )

        new_position = self.safe_transition(position=self.current_position, move=move)

        self.last_change = StateChange(
            start_position=self.current_position,
            reward_change=reward,
            was_covered=self.uncovered[new_position],
        )

        self.uncovered[new_position] = True
        self.accumulated_reward += reward
        self.current_position = new_position
        self.moves += 1

        self.cached_transition.stop()

        return reward

    def undo(self) -> int:
        """Undoes the last move made in the grid game.

        Returns:
            The reward for undoing the last move.
        """
        assert (change := self.last_change) is not None, "No moves have been made yet."

        self.uncovered[self.current_position] = change.was_covered
        self.accumulated_reward -= change.reward_change
        self.current_position = change.start_position
        self.moves -= 1

        return -change.reward_change

    def reset(self) -> "GridGame":
        """Creates a new grid game starting fresh from the initial position."""
        if self.policy is not None:
            self.policy.reset()

        return GridGame(
            safe_transition=self.safe_transition,
            cached_transition=self.cached_transition,
            original_transition=self.original_transition,
            reward_function=self.reward_function,
            rewards=self.rewards,
            uncovered=mask_for(self.rewards, self.start_position),
            start_position=self.start_position,
            solver=self.solver,
            policy=self.policy,
            current_position=self.start_position,
            solution=self.solution,
        )

    def regenerate(self) -> "GridGame":
        """Creates a new grid game with different rewards."""
        return GridGame.create(
            rows=self.rows,
            columns=self.columns,
            reward_function=self.reward_function,
            transition_function=self.original_transition,
            reward_range=(self.rewards.min(), self.rewards.max()),
            start_position=self.start_position,
            solver=self.solver,
            policy=self.policy,
        )

    def with_solution(self, solution: Solution) -> "GridGame":
        """Creates a new grid game that has an optimal solution."""
        return GridGame(
            safe_transition=self.safe_transition,
            cached_transition=self.cached_transition,
            original_transition=self.original_transition,
            reward_function=self.reward_function,
            rewards=self.rewards,
            uncovered=self.uncovered,
            start_position=self.start_position,
            solver=self.solver,
            policy=self.policy,
            current_position=self.current_position,
            solution=solution,
        )

    def optimal_solution(self) -> Solution:
        """Returns the optimal solution to the grid game."""
        assert self.solution is not None, (
            "The optimal solution has not been computed yet."
        )

        return self.solution

    @property
    def rows(self) -> int:
        """The number of rows in the grid."""
        return self.rewards.shape[0]

    @property
    def columns(self) -> int:
        """The number of columns in the grid."""
        return self.rewards.shape[1]

    @property
    def all_positions(self) -> list[Position]:
        """Returns all valid positions in the grid."""
        return [Position(r, c) for r in range(self.rows) for c in range(self.columns)]

    @property
    def last_reward(self) -> int:
        """The reward for the last move made."""
        return self.last_change.reward_change if self.last_change else 0


def random_rewards(
    *, rows: int, columns: int, reward_range: tuple[int, int] = (-5, 5)
) -> IntMatrix:
    min_rewards, max_reward = reward_range

    # Extra shift to get more negative rewards
    rewards = (
        np.random.randint(*reward_range, size=(rows, columns)) - abs(max_reward) // 2
    )

    # Add pattern: higher rewards along edges and from center to right
    extra_rewards = abs(max_reward) // 2, 2 * abs(max_reward)

    rewards[0, :] += np.random.randint(*extra_rewards, size=columns)
    rewards[rows - 1, :] += np.random.randint(*extra_rewards, size=columns)
    rewards[:, 0] += np.random.randint(*extra_rewards, size=rows)
    rewards[:, columns - 1] += np.random.randint(*extra_rewards, size=rows)
    rewards[rows // 2, columns // 2 :] += np.random.randint(
        *extra_rewards, size=columns - columns // 2
    )

    return np.clip(rewards, min_rewards, max_reward)


def mask_for(grid: IntMatrix, start_position: Position) -> BoolMatrix:
    mask = np.zeros_like(grid, dtype=bool)
    mask[start_position] = True
    return mask



from typing import Protocol, Any, TypeVar, Generic, TypeAlias
from helpers.network.scaler import InputScaler, OutputScaler

import os
import torch
import msgspec.json as json
import joblib


ConfigurationT = TypeVar("ConfigurationT")
NetworkT = TypeVar("NetworkT", covariant=True)


StateDict: TypeAlias = dict[str, Any]


class NeuralNetwork(Protocol):
    def configuration(self) -> Any:
        """Returns the parameters of the network."""
        ...

    def state(self) -> StateDict:
        """Returns the state of the network."""
        ...

    def input_scaler(self) -> InputScaler:
        """Returns the input scaler of the network."""
        ...

    def output_scaler(self) -> OutputScaler:
        """Returns the output scaler of the network."""
        ...


class NeuralNetworkCreator(Protocol, Generic[ConfigurationT, NetworkT]):
    def create(
        self,
        configuration: ConfigurationT,
        state: StateDict,
        *,
        input_scaler: InputScaler,
        output_scaler: OutputScaler,
    ) -> NetworkT:
        """Creates a new network with the given configuration and state."""
        ...

    def configuration_type(self) -> type[ConfigurationT]:
        """Returns the type of the configuration."""
        ...


def save(network: NeuralNetwork, *, to: str) -> None:
    os.makedirs(to, exist_ok=True)
    remove_check_file(to)

    torch.save(network.state(), f"{to}/state.pth")
    joblib.dump(network.input_scaler(), f"{to}/input-scaler.joblib")
    joblib.dump(network.output_scaler(), f"{to}/output-scaler.joblib")

    with open(f"{to}/configuration.json", "wb") as file:
        file.write(json.encode(network.configuration()))

    create_check_file(to)


def load(
    path: str, *, using: NeuralNetworkCreator[ConfigurationT, NetworkT]
) -> NetworkT:
    assert os.path.exists(path), (
        f"The directory {path} containing the network does not exist."
    )

    check_check_file(path)

    state = torch.load(f"{path}/state.pth", weights_only=True)
    input_scaler = joblib.load(f"{path}/input-scaler.joblib")
    output_scaler = joblib.load(f"{path}/output-scaler.joblib")

    with open(f"{path}/configuration.json", "rb") as file:
        parameters = json.decode(file.read(), type=using.configuration_type())

    return using.create(
        parameters, state, input_scaler=input_scaler, output_scaler=output_scaler
    )


def remove_check_file(path: str) -> None:
    try:
        os.remove(f"{path}/serialization.complete")
    except FileNotFoundError:
        pass


def create_check_file(path: str) -> None:
    with open(f"{path}/serialization.complete", "w") as file:
        file.write(
            "This file is used to check if the serialization completed successfully."
        )


def check_check_file(path: str) -> None:
    if not os.path.exists(f"{path}/serialization.complete"):
        raise RuntimeError(
            "Looks like the model was not serialized properly. "
            "Unfortunately, you can't reliably load it."
        )



from dataclasses import dataclass, KW_ONLY

from helpers.maths import Vector


@dataclass(frozen=True)
class TankStates:
    """The states of a tank system over a period of time.

    Attributes:
        time: The time points.
        inlet_flow_rate: The inlet flow rates.
        liquid_height: The liquid heights.
        outlet_flow_rate: The outlet flow rates.
    """

    _: KW_ONLY
    time: Vector
    inlet_flow_rate: Vector
    liquid_height: Vector
    outlet_flow_rate: Vector

    def __post_init__(self) -> None:
        assert (
            len(self.time)
            == len(self.inlet_flow_rate)
            == len(self.liquid_height)
            == len(self.outlet_flow_rate)
        ), (
            f"Expected all vectors to have the same length."
            f"Got time: {len(self.time)}, inlet_flow_rate: {len(self.inlet_flow_rate)}, "
            f"liquid_height: {len(self.liquid_height)}, outlet_flow_rate: {len(self.outlet_flow_rate)}."
        )

    def until(self, index: int) -> "TankStates":
        return TankStates(
            time=self.time[:index],
            inlet_flow_rate=self.inlet_flow_rate[:index],
            liquid_height=self.liquid_height[:index],
            outlet_flow_rate=self.outlet_flow_rate[:index],
        )

    @property
    def time_range(self) -> tuple[float, float]:
        return self.time[0], self.time[-1]

    @property
    def flow_rate_range(self) -> tuple[float, float]:
        return (
            min(self.inlet_flow_rate.min(), self.outlet_flow_rate.min()),
            max(self.inlet_flow_rate.max(), self.outlet_flow_rate.max()),
        )

    @property
    def points(self) -> int:
        return len(self.time)



from .data import DataLoader as DataLoader, DataMapper as DataMapper
from .scaler import InputScaler as InputScaler, OutputScaler as OutputScaler
from .serialize import save as save, load as load, StateDict as StateDict
from .info import describe as describe
from .training import (
    Data as Data,
    BatchLoader as BatchLoader,
    TrainingVisualizer as TrainingVisualizer,
)
from .visualize import (
    ActivationLayerVisualizerMixin as ActivationLayerVisualizerMixin,
    LossFunctionVisualizerMixin as LossFunctionVisualizerMixin,
    visualize_loss as visualize_loss,
)



from .types import (
    DynamicsSolver as DynamicsSolver,
    RobotDynamics as RobotDynamics,
    State as State,
    Position as Position,
    Jacobian as Jacobian,
    Gravity as Gravity,
    ControlSignal as ControlSignal,
    StatefulControlSignal as StatefulControlSignal,
)
from .simulate import (
    RobotArmSimulationMixin as RobotArmSimulationMixin,
)
from .results import (
    SimulationResults as SimulationResults,
    SimulatableRobotArm as SimulatableRobotArm,
    AnimatedRobotArm as AnimatedRobotArm,
)
from .common import (
    PrecomputedState as PrecomputedState,
    LazyPosition as LazyPosition,
    LazyJacobian as LazyJacobian,
    PrecomputedGravity as PrecomputedGravity,
    UnknownGravity as UnknownGravity,
)
from .measurement import (
    infer as infer,
    visualize as visualize,
    MeasurementColumns as MeasurementColumns,
)
from .solver import EulerSolver as EulerSolver, LsodaSolver as LsodaSolver



from typing import Sequence

from helpers.maths import Matrix, Vector
from helpers.robot.arm import RobotArm

import numpy as np


class RobotJacobianMixin:
    def jacobian(self: RobotArm, q: Sequence[float] | Vector) -> Matrix:
        """Returns the Jacobian matrix of the robot arm at the given configuration.

        Args:
            q: The configuration of the robot arm joints.

        Returns:
            The Jacobian matrix of the robot arm. The matrix is of shape (2, 2).
        
        Note:
            The Jacobian matrix maps the joint velocities to the end effector velocities. We assume
            that the robot has two revolute joints and the forward kinematics are given by:

            $$
            x = l_1 cos(\\theta_1) + l_2 cos(\\theta_1 + \\theta_2) \\
            y = l_1 sin(\\theta_1) + l_2 sin(\\theta_1 + \\theta_2)
            $$
            
            The Jacobian matrix is then given by:

            $$
            J = \\begin{bmatrix}
            -l_1 sin(\\theta_1) - l_2 sin(\\theta_1 + \\theta_2) & -l_2 sin(\\theta_1 + \\theta_2) \\
            l_1 cos(\\theta_1) + l_2 cos(\\theta_1 + \\theta_2) & l_2 cos(\\theta_1 + \\theta_2)
            \\end{bmatrix}
            $$
        """

        assert len(q) == 2, (
            f"The configuration must contain exactly two joint angles. Got: {len(q)}"
        )

        q_1, q_2 = q
        return np.array(
            [
                [
                    -self.l_1 * np.sin(q_1) - self.l_2 * np.sin(q_1 + q_2),
                    -self.l_2 * np.sin(q_1 + q_2),
                ],
                [
                    self.l_1 * np.cos(q_1) + self.l_2 * np.cos(q_1 + q_2),
                    self.l_2 * np.cos(q_1 + q_2),
                ],
            ]
        )



from typing import Sequence
from dataclasses import dataclass, KW_ONLY
from helpers.maths import Vector
from helpers.robot.control.control import (
    SetpointFollower,
    TrajectoryFollower,
    Trajectory,
)
from helpers.robot.simulate import (
    ControlSignal,
    State,
    Position,
)

from numpy.linalg import norm


def task_space_distance(*, q: State, p: Position, t: float, target: Vector) -> float:
    return float(norm(p(q) - target))


# TODO: Untested!
@dataclass(frozen=True)
class PD:
    _: KW_ONLY
    K_x_p: float
    K_x_d: float
    gravity_compensation: bool

    @staticmethod
    def with_parameters(
        *, w_n: float, xi: float, M_x: float, gravity_compensation: bool = False
    ) -> "PD":
        assert w_n > 0, f"The natural frequency must be greater than zero. Got {w_n}."
        assert 0 < xi <= 1, f"The damping ratio must be between 0 and 1. Got {xi}."
        assert M_x > 0, (
            f"The guess for the inertia of the system must be greater than zero. Got {M_x}."
        )

        return PD(
            K_x_p=w_n**2 * M_x,
            K_x_d=2 * xi * w_n * M_x,
            gravity_compensation=gravity_compensation,
        )

    def setpoint(self, p_d: Vector) -> ControlSignal:
        if self.gravity_compensation:
            return lambda q, p, J, g, t: -J(q).T @ (
                self.K_x_p * (p(q) - p_d) + self.K_x_d * J(q) @ q.dot(t) + g(q)
            )
        else:
            return lambda q, p, J, g, t: -J(q).T @ (
                self.K_x_p * (p(q) - p_d) + self.K_x_d * J(q) @ q.dot(t)
            )

    def setpoints(
        self, p_d: Sequence[Vector], *, threshold: float = 0.1
    ) -> ControlSignal:
        return SetpointFollower(
            setpoints=p_d,
            control_provider=self.setpoint,
            distance=task_space_distance,
            threshold=threshold,
        )

    def trajectory(self, p_d: Trajectory) -> ControlSignal:
        return TrajectoryFollower(trajectory=p_d, control_provider=self.setpoint)



from typing import Iterator
from dataclasses import dataclass, field

from helpers.maths import Matrix

from dash import Dash, html, dcc
from dash.dependencies import Input, Output
import dash_bootstrap_components as dbc

from plotly.graph_objs import Figure
from torch.utils.data import DataLoader, TensorDataset
from torch import Tensor


@dataclass(frozen=True)
class BatchLoader:
    loader: DataLoader

    @property
    def element_count(self) -> int:
        return len(self.loader.dataset)  # type: ignore

    @property
    def batch_count(self) -> int:
        return len(self.loader)

    def __iter__(self) -> Iterator[tuple[Tensor, Tensor]]:
        yield from self.loader

    def __len__(self) -> int:
        return self.batch_count


@dataclass(frozen=True)
class Data:
    train: tuple[Matrix, Matrix]
    val: tuple[Matrix, Matrix]
    test: tuple[Matrix, Matrix]

    @staticmethod
    def partitions(
        train: tuple[Matrix, Matrix],
        val: tuple[Matrix, Matrix],
        test: tuple[Matrix, Matrix],
    ) -> "Data":
        return Data(train=train, val=val, test=test)

    def training(self, *, batch_size: int) -> BatchLoader:
        return self._data_loader_for(self.train, batch_size=batch_size)

    def validation(self, *, batch_size: int) -> BatchLoader:
        return self._data_loader_for(self.val, batch_size=batch_size)

    def testing(self, *, batch_size: int) -> BatchLoader:
        return self._data_loader_for(self.test, batch_size=batch_size)

    def _data_loader_for(
        self, data_partition: tuple[Matrix, Matrix], *, batch_size: int
    ) -> BatchLoader:
        X, y = data_partition
        X_tensor = Tensor(X).float()
        y_tensor = Tensor(y).float()

        dataset = TensorDataset(X_tensor, y_tensor)
        return BatchLoader(DataLoader(dataset, batch_size=batch_size, shuffle=True))


@dataclass
class TrainingVisualizer:
    total_epochs: int

    last_epoch: int = field(init=False, default=0)
    training_losses: list[float] = field(init=False, default_factory=list)
    val_losses: list[float] = field(init=False, default_factory=list)

    app: Dash = field(init=False)

    def __post_init__(self) -> None:
        self.app = Dash(__name__, external_stylesheets=[dbc.themes.FLATLY])
        self._create_layout_for(self.app)
        self._setup_callbacks_for(self.app)

    def start(self, *, height: int = 450) -> "TrainingVisualizer":
        self.app.run(jupyter_height=height)
        return self

    def complete(self, *, epoch: int, training_loss: float, val_loss: float) -> None:
        self.last_epoch = epoch
        self.training_losses.append(training_loss)
        self.val_losses.append(val_loss)

    def visualize(self) -> Figure:
        return Figure(
            data=[
                {
                    "x": list(range(1, self.last_epoch + 1)),
                    "y": self.training_losses,
                    "name": "Training Loss",
                },
                {
                    "x": list(range(1, self.last_epoch + 1)),
                    "y": self.val_losses,
                    "name": "Validation Loss",
                },
                {
                    "x": [self.last_epoch, self.last_epoch],
                    "y": [
                        0,
                        max(max(self.training_losses), max(self.val_losses))
                        if self.last_epoch > 0
                        else 1,
                    ],
                    "mode": "lines",
                    "line": {"color": "black", "width": 2, "dash": "dash"},
                    "name": f"Epoch {self.last_epoch}",
                },
            ],
            layout={
                "title": "Training and Validation Loss",
                "xaxis": {"title": "Epoch", "range": [1, self.total_epochs]},
                "yaxis": {"title": "Loss"},
            },
        )

    def _create_layout_for(self, app: Dash) -> None:
        app.layout = dbc.Container(
            [
                dbc.Row(
                    [
                        dbc.Col(
                            html.Div(
                                [
                                    dcc.Graph(id="loss-graph", figure=self.visualize()),
                                    dcc.Interval(
                                        id="interval-component",
                                        interval=500,  # in milliseconds
                                        n_intervals=0,
                                    ),
                                ]
                            )
                        )
                    ]
                )
            ]
        )

    def _setup_callbacks_for(self, app: Dash) -> None:
        @app.callback(
            Output("loss-graph", "figure"),
            Input("interval-component", "n_intervals"),
        )
        def update_graph(n: int) -> Figure:
            return self.visualize()



from typing import Protocol, TypeVar, Generic, Sequence

StateT = TypeVar("StateT", bound="State", infer_variance=True)
ActionT = TypeVar("ActionT", infer_variance=True)


class State(Protocol):
    @property
    def value(self) -> float:
        """Returns the value of the state."""
        ...


class RewardFunction(Protocol, Generic[StateT]):
    def __call__(self, state: StateT, /) -> float:
        """Returns the reward for the given state."""
        ...


class Policy(Protocol, Generic[ActionT]):
    def __call__(self, actions: Sequence[tuple[ActionT, float]], /) -> ActionT:
        """Returns the action to take based on the policy."""
        ...



from typing import Final, NamedTuple, TypeAlias
from enum import Enum

from helpers.rl.game import (
    GridGame,
    RewardFunction,
    TransitionFunction,
    Policy,
    DeterministicTransition,
    Moves,
    EndGame,
)
from helpers.rl.animate import GridGameVisualizer

import dash
from dash import Dash, html, dcc
from dash.dependencies import Input, Output
from dash.development.base_component import Component
import dash_bootstrap_components as dbc

from plotly.graph_objects import Figure


Content: TypeAlias = str | list[str | Component] | Component


class PlayerActionItems(str, Enum):
    RESTART = "restart"
    END_GAME = "end-game"
    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"


class AgentActionItems(str, Enum):
    START = "start"
    STOP = "stop"
    INTERVAL = "interval"


class PlayerUpdateGame(NamedTuple):
    figure: Figure
    score_display: Content
    score_color: str


class AgentUpdateGame(NamedTuple):
    figure: Figure
    score_display: Content
    score_color: str
    stop_disabled: bool
    interval_disabled: bool

    @staticmethod
    def stopped(
        figure: Figure, score_display: Content, score_color: str
    ) -> "AgentUpdateGame":
        return AgentUpdateGame(
            figure,
            score_display,
            score_color,
            stop_disabled=True,
            interval_disabled=True,
        )

    @staticmethod
    def running(
        figure: Figure, score_display: Content, score_color: str
    ) -> "AgentUpdateGame":
        return AgentUpdateGame(
            figure,
            score_display,
            score_color,
            stop_disabled=False,
            interval_disabled=False,
        )


class UpdateButtons(NamedTuple):
    up_disabled: bool
    down_disabled: bool
    left_disabled: bool
    right_disabled: bool

    @staticmethod
    def all_enabled() -> "UpdateButtons":
        return UpdateButtons(False, False, False, False)

    @staticmethod
    def all_disabled() -> "UpdateButtons":
        return UpdateButtons(True, True, True, True)


class GridGameApp:
    visualizer: GridGameVisualizer
    figure: Figure
    app: Dash
    game: GridGame

    MOVE_MAP: Final = {
        PlayerActionItems.UP: Moves.UP,
        PlayerActionItems.DOWN: Moves.DOWN,
        PlayerActionItems.LEFT: Moves.LEFT,
        PlayerActionItems.RIGHT: Moves.RIGHT,
    }

    @staticmethod
    def create(
        *,
        rows: int,
        columns: int,
        reward_function: RewardFunction,
        transition_function: TransitionFunction = DeterministicTransition(),
        policy: Policy | None = None,
        move_interval_ms: int = 500,
    ) -> "GridGameApp":
        """Create a new grid game application.

        Args:
            rows: Number of rows in the grid.
            columns: Number of columns in the grid.
            reward_function: A reward function for the game.
            transition_function: A transition function for the game.
            policy: A policy to use for the game. If not provided, the user will play the game.
            move_interval_ms: The interval in milliseconds between moves for the agent
                (only applicable if a policy is provided).

        Returns:
            A GridGameApp instance.

        Example:
            You can start a new game with 5 rows and 5 columns as follows:

            ```python
            app = GridGameApp.create(rows=5, columns=5, reward_function=MyAwesomeRewardFunction())
            app.run()
            ```
        """
        return GridGameApp(
            GridGame.create(
                rows=rows,
                columns=columns,
                reward_function=reward_function,
                transition_function=transition_function,
                policy=policy,
            ),
            move_interval_ms=move_interval_ms,
        )

    def __init__(self, game: GridGame, *, move_interval_ms: int) -> None:
        self.app = Dash(__name__, external_stylesheets=[dbc.themes.FLATLY])

        self._initialize(game)
        self._create_layout(self.app, self.figure, move_interval_ms=move_interval_ms)
        self._setup_callbacks()

    def run(self, debug: bool = False, port: str = "8050") -> None:
        self.app.run(debug=debug, port=port, jupyter_height=400, jupyter_width="500")

    def _initialize(self, game: GridGame) -> None:
        self.game = game
        self.visualizer = GridGameVisualizer(game)
        self.figure = self.visualizer.create()

    def _create_layout(
        self, app: Dash, figure: Figure, *, move_interval_ms: int
    ) -> None:
        app.layout = dbc.Container(
            [
                dbc.Row(
                    [
                        dbc.Col(
                            [
                                html.H1("Grid Game", className="text-center my-2"),
                                self._score_display(),
                                self._result_area(),
                            ],
                            style={"maxWidth": "500px"},
                        ),
                    ],
                    align="top",
                    justify="center",
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            [self._game_board(figure)], style={"maxWidth": "250px"}
                        ),
                        dbc.Col(
                            [self._controls(move_interval_ms=move_interval_ms)],
                            style={"maxWidth": "250px"},
                        ),
                    ],
                    align="top",
                    justify="center",
                ),
            ],
            fluid=True,
            className="py-2",
        )

    def _score_display(self) -> dbc.Alert:
        return dbc.Alert(
            id="score-display",
            color="light",
            className="text-center mb-4",
        )

    def _game_board(self, figure: Figure) -> dcc.Graph:
        return dcc.Graph(
            id="game-board",
            figure=figure,
            config={"displayModeBar": False},
        )

    def _controls(self, *, move_interval_ms: int) -> dbc.Card:
        return (
            self._player_controls()
            if self.game.policy is None
            else self._agent_controls(move_interval_ms=move_interval_ms)
        )

    def _player_controls(self) -> dbc.Card:
        return dbc.Card(
            [
                dbc.CardBody(
                    [
                        self._player_explanation(),
                        self._direction_pad(),
                        html.Div(self._player_game_menu(), className="mt-auto"),
                    ],
                    className="d-flex flex-column h-100",
                )
            ],
            className="h-100",
        )

    def _agent_controls(self, *, move_interval_ms: int) -> dbc.Card:
        return dbc.Card(
            [
                dbc.CardBody(
                    [
                        self._agent_explanation(),
                        html.Div(self._agent_game_menu(), className="mt-auto"),
                        dcc.Interval(
                            id=AgentActionItems.INTERVAL,
                            interval=move_interval_ms,
                            n_intervals=0,
                            disabled=True,
                        ),
                    ],
                    className="d-flex flex-column h-100",
                )
            ],
            className="h-100",
        )

    def _player_explanation(self) -> html.Div:
        return html.Div(
            [
                html.P([html.Strong("• Movements: "), "↑ ↓ ← →"], className="mb-0"),
                html.P([html.Strong("• Player: "), "White Dot"], className="mb-0"),
                html.Hr(className="my-2"),
            ],
            className="text-start small",
            style={"fontSize": "12px", "padding": "0 0 5px 0"},
        )

    def _agent_explanation(self) -> html.Div:
        return html.Div(
            [html.Strong("Press Start to begin.")],
            className="text-center small",
            style={"fontSize": "12px", "padding": "0 0 5px 0"},
        )

    def _direction_pad(self, *, element_width: int = 2) -> dbc.Container:
        button_style = {
            "width": "30px",
            "height": "30px",
            "padding": "0px",
        }

        return dbc.Container(
            [
                dbc.Row(
                    [
                        dbc.Col(width=element_width),
                        dbc.Col(
                            dbc.Button(
                                "↑",
                                id=PlayerActionItems.UP,
                                color="primary",
                                n_clicks=0,
                                style=button_style,
                            ),
                            width=element_width,
                            className="text-center p-0",
                        ),
                        dbc.Col(width=element_width),
                    ],
                    className="g-0",
                    justify="center",
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            dbc.Button(
                                "←",
                                id=PlayerActionItems.LEFT,
                                color="primary",
                                n_clicks=0,
                                style=button_style,
                            ),
                            width=element_width,
                            className="text-center p-0",
                        ),
                        dbc.Col(width=element_width),
                        dbc.Col(
                            dbc.Button(
                                "→",
                                id=PlayerActionItems.RIGHT,
                                color="primary",
                                n_clicks=0,
                                style=button_style,
                            ),
                            width=element_width,
                            className="text-center p-0",
                        ),
                    ],
                    className="g-0 my-1",
                    justify="center",
                ),
                dbc.Row(
                    [
                        dbc.Col(width=element_width),
                        dbc.Col(
                            dbc.Button(
                                "↓",
                                id=PlayerActionItems.DOWN,
                                color="primary",
                                n_clicks=0,
                                style=button_style,
                            ),
                            width=element_width,
                            className="text-center p-0",
                        ),
                        dbc.Col(width=element_width),
                    ],
                    className="g-0",
                    justify="center",
                ),
            ],
            fluid=True,
            className="p-0",
        )

    def _player_game_menu(self) -> dbc.Row:
        button_style = {"width": "80px", "margin": "0", "font-size": "0.7rem"}

        return dbc.Row(
            [
                dbc.Col(
                    [
                        dbc.Button(
                            "Restart",
                            id=PlayerActionItems.RESTART,
                            color="secondary",
                            n_clicks=0,
                            style=button_style,
                        ),
                    ],
                    className="text-center mt-2",
                ),
                dbc.Col(
                    [
                        dbc.Button(
                            "End Game",
                            id=PlayerActionItems.END_GAME,
                            color="info",
                            n_clicks=0,
                            style=button_style,
                        ),
                    ],
                    className="text-center mt-2",
                ),
            ]
        )

    def _agent_game_menu(self) -> dbc.Row:
        button_style = {"width": "80px", "margin": "0 5px", "font-size": "0.7rem"}

        return dbc.Row(
            [
                dbc.Col(
                    [
                        dbc.Button(
                            "Start",
                            id=AgentActionItems.START,
                            color="primary",
                            n_clicks=0,
                            style=button_style,
                        ),
                        dbc.Button(
                            "Stop",
                            id=AgentActionItems.STOP,
                            color="light",
                            n_clicks=0,
                            style=button_style,
                        ),
                    ],
                    className="text-center mt-2",
                ),
            ]
        )

    def _result_area(self) -> dbc.Alert:
        return dbc.Alert(
            id="message-area",
            color="info",
            is_open=False,
            duration=None,
            className="mt-4",
        )

    def _setup_callbacks(self) -> None:
        if self.game.policy is None:
            self._setup_player_callbacks()
        else:
            self._setup_agent_callbacks()

    def _setup_player_callbacks(self) -> None:
        @self.app.callback(
            Output("game-board", "figure"),
            Output("score-display", "children"),
            Output("score-display", "color"),
            [
                Input(PlayerActionItems.UP, "n_clicks"),
                Input(PlayerActionItems.DOWN, "n_clicks"),
                Input(PlayerActionItems.LEFT, "n_clicks"),
                Input(PlayerActionItems.RIGHT, "n_clicks"),
                Input(PlayerActionItems.RESTART, "n_clicks"),
                Input(PlayerActionItems.END_GAME, "n_clicks"),
            ],
        )
        def update_game(up, down, left, right, restart, end_game) -> PlayerUpdateGame:
            ctx = dash.callback_context
            if not ctx.triggered:
                return PlayerUpdateGame(
                    self.figure, self._score_info(self.game), "light"
                )

            button_id = ctx.triggered[0]["prop_id"].split(".")[0]

            match button_id:
                case PlayerActionItems.RESTART:
                    self._initialize(self.game.regenerate())
                    return PlayerUpdateGame(
                        self.figure, self._score_info(self.game), "light"
                    )
                case PlayerActionItems.END_GAME:
                    solution = self.game.optimal_solution()
                    return PlayerUpdateGame(
                        self.figure,
                        (
                            f"Game Over! Your score: {self.game.accumulated_reward} | "
                            f"Our high score: {solution.reward}"
                        ),
                        "info",
                    )
                case _:
                    assert button_id in GridGameApp.MOVE_MAP, (
                        f"Unknown button: {button_id}"
                    )

                    self.game.move(GridGameApp.MOVE_MAP[button_id])

                    return PlayerUpdateGame(
                        self.visualizer.update(self.figure),
                        self._score_info(self.game),
                        "light",
                    )

        @self.app.callback(
            [
                Output(PlayerActionItems.UP, "disabled"),
                Output(PlayerActionItems.DOWN, "disabled"),
                Output(PlayerActionItems.LEFT, "disabled"),
                Output(PlayerActionItems.RIGHT, "disabled"),
            ],
            [
                Input(PlayerActionItems.RESTART, "n_clicks"),
                Input(PlayerActionItems.END_GAME, "n_clicks"),
            ],
        )
        def update_buttons(restart_clicks, end_clicks) -> UpdateButtons:
            ctx = dash.callback_context
            if not ctx.triggered:
                return UpdateButtons.all_enabled()

            button_id = ctx.triggered[0]["prop_id"].split(".")[0]

            match button_id:
                case PlayerActionItems.END_GAME:
                    return UpdateButtons.all_disabled()
                case _:
                    return UpdateButtons.all_enabled()

    def _setup_agent_callbacks(self) -> None:
        @self.app.callback(
            Output("game-board", "figure"),
            Output("score-display", "children"),
            Output("score-display", "color"),
            Output(AgentActionItems.INTERVAL, "disabled"),
            Output(AgentActionItems.STOP, "disabled"),
            [
                Input(AgentActionItems.INTERVAL, "n_intervals"),
                Input(AgentActionItems.START, "n_clicks"),
                Input(AgentActionItems.STOP, "n_clicks"),
            ],
        )
        def update_agent_game(
            n_intervals, start_clicks, stop_clicks
        ) -> AgentUpdateGame:
            ctx = dash.callback_context
            if not ctx.triggered:
                return AgentUpdateGame.stopped(
                    self.figure, self._score_info(self.game), "light"
                )

            button_id = ctx.triggered[0]["prop_id"].split(".")[0]

            match button_id:
                case AgentActionItems.START:
                    self._initialize(self.game.regenerate())
                    return AgentUpdateGame.running(
                        self.figure, self._score_info(self.game), "light"
                    )
                case AgentActionItems.STOP:
                    return AgentUpdateGame.stopped(
                        self.figure, self._score_info(self.game), "light"
                    )
                case _:
                    try:
                        self.game.move()
                        return AgentUpdateGame.running(
                            self.visualizer.update(self.figure),
                            self._score_info(self.game),
                            "light",
                        )
                    except EndGame:
                        solution = self.game.optimal_solution()
                        return AgentUpdateGame.stopped(
                            self.visualizer.update(self.figure),
                            (
                                f"Game Over! Your score: {self.game.accumulated_reward} | "
                                f"Our high score: {solution.reward}"
                            ),
                            "info",
                        )

    def _score_info(self, game: GridGame) -> Component:
        accumulated = game.accumulated_reward
        last = game.last_reward

        last_reward_color = "green" if last > 0 else "red" if last < 0 else "inherit"

        return html.Div(
            [
                html.P(f"Score: {accumulated}", className="mb-0 me-4"),
                html.P(
                    [
                        "Last Reward: ",
                        html.Span(f"{last}", style={"color": last_reward_color}),
                    ],
                    className="mb-0",
                ),
            ],
            className="d-flex",
            style={"justifyContent": "space-around", "alignItems": "center"},
        )



from dataclasses import dataclass, KW_ONLY


@dataclass(frozen=True)
class Link:
    """A link in a robot arm.

    The link is assumed to be a rigid body with a center of mass somewhere along its length. The
    center of mass is assumed to be a cube with a side length of `com_size`. The remaining part of
    the link is assumed to be a thin rod (i.e. massless).

    Attributes:
        mass: The mass of the link.
        length: The length of the link.
        distance_to_com: The distance from the base of the link to its center of mass.
        com_size: The size of the center of mass (defaults to 0.1).

    Example:
        When creating a [`PlanarRobotArm`](exercise_3.md#helpers.robot.PlanarRobotArm), you can define
        the links of the robot arm as follows:

        ```python
        robot = PlanarRobotArm(
            link_1=Link(mass=1.0, length=1.0, distance_to_com=0.75),
            link_2=Link(mass=2.0, length=0.75, distance_to_com=0.5),
            ... # Other parameters
        )

        # This will create a planar robot arm with two links.
        ```
    """

    _: KW_ONLY
    mass: float
    length: float
    distance_to_com: float
    com_size: float = 0.1

    def __post_init__(self) -> None:
        assert self.mass > 0, (
            f"The mass of the link must be greater than zero. Got {self.mass}."
        )
        assert self.length > 0, (
            f"The length of the link must be greater than zero. Got {self.length}."
        )
        assert 0 <= self.distance_to_com <= self.length, (
            f"The center of mass must be located within the link, "
            f"i.e. the value must be between 0 and {self.length}. Got {self.distance_to_com}."
        )
        assert self.com_size > 0, (
            f"The size of the center of mass must be greater than zero. Got {self.com_size}."
        )



from dataclasses import dataclass, KW_ONLY

from helpers.tank.animation import VisualizeTankMixin
from helpers.tank.simulate import SimulateTankMixin
from helpers.tank.dynamics import TankDynamicsMixin
from helpers.tank.discretize import DiscretizeTankMixin

from math import pi, sqrt


@dataclass
class Tank(
    VisualizeTankMixin, SimulateTankMixin, TankDynamicsMixin, DiscretizeTankMixin
):
    name: str
    _: KW_ONLY
    height: float
    diameter: float
    outlet_constant: float
    max_inlet_flow: float
    liquid_height: float

    @staticmethod
    def create(
        name: str,
        *,
        height: float,
        diameter: float,
        outlet_constant: float,
        max_inlet_flow: float,
        liquid_height: float = 0,
    ) -> "Tank":
        return Tank(
            name,
            height=height,
            diameter=diameter,
            outlet_constant=outlet_constant,
            max_inlet_flow=max_inlet_flow,
            liquid_height=liquid_height,
        )

    def liquid_height_is(self, height: float) -> "Tank":
        self.liquid_height = height
        return self

    def inlet_flow_rate_for(self, control: float) -> float:
        return max(0, min(control, self.max_inlet_flow))

    def outlet_flow_rate_for(self, height: float) -> float:
        return self.outlet_constant * sqrt(height)

    @property
    def area(self) -> float:
        return pi * (self.diameter / 2) ** 2



from typing import Protocol, Sequence
from helpers.maths import Vector
from plotly.graph_objects import Figure, Scatter, Scatter3d


import numpy as np


class ActivationLayer(Protocol):
    def forward(self, sigma: Vector, /) -> Vector:
        """Compute the forward pass of the activation layer."""
        ...


class LossFunction(Protocol):
    def loss(self, *, y_true: Vector, y_predicted: Vector) -> float:
        """Compute the loss of the loss function."""
        ...


class ActivationLayerVisualizerMixin:
    def visualize(
        self: ActivationLayer,
        *,
        sigma_range: tuple[float, float] = (-5, 5),
        points: int = 100,
    ) -> None:
        """Visualize the activation layer for a one-dimensional input.

        Args:
            sigma_range: The range of sigma values to visualize.
            points: The number of points to visualize.

        Example:
            ```python
            layer = SigmoidLayer()
            layer.visualize()
            # Output: A plot of the sigmoid function.
            ```
        """
        sigma = np.linspace(*sigma_range, points)
        a = self.forward(sigma)

        figure = Figure(data=[Scatter(x=sigma, y=a, name="Activation")])
        figure.update_layout(title=f"{self.__class__.__name__} Activation Layer")
        figure.show()


class LossFunctionVisualizerMixin:
    def visualize(
        self: LossFunction,
        *,
        y_range: tuple[float, float] = (-5, 5),
        points: int = 10,
        classification: bool = False,
    ) -> None:
        """Visualize the loss function for a one-dimensional input in two dimensions.

        Args:
            y_range: The range of y values to visualize.
            points: The number of points to visualize.
            classification: Whether the loss function is for classification.

        Example:
            ```python
            loss = MeanSquaredError()
            loss.visualize()
            # Output: A plot of the mean squared error.
            ```
        """

        if classification:
            _visualize_classification(loss=self, points=points)
        else:
            _visualize_regression(loss=self, y_range=y_range, points=points)


def visualize_loss(losses: Sequence[float]) -> None:
    """Visualize a sequence of computed losses.

    Args:
        losses: A sequence of computed losses.

    Example:
        ```python
        losses = [0.5, 0.4, 0.3, 0.2, 0.1]
        visualize_loss(losses)
        # Output: A plot of the losses.
        ```
    """

    figure = Figure(data=[Scatter(y=losses, name="Loss")])
    figure.update_layout(title="Losses")
    figure.show()


def _visualize_regression(
    loss: LossFunction, *, y_range: tuple[float, float], points: int
) -> None:
    # Create a grid of y values
    y_true = np.linspace(*y_range, points)
    y_predicted = np.linspace(*y_range, points)
    Y_true, Y_predicted = np.meshgrid(y_true, y_predicted)

    # Compute the loss for each pair of y values
    Z = np.zeros_like(Y_true)
    for i, y_t in enumerate(y_true):
        for j, y_p in enumerate(y_predicted):
            Z[i, j] = loss.loss(y_true=y_t, y_predicted=y_p)

    # Create a 3D surface plot
    figure = Figure(
        data=[
            Scatter3d(
                x=Y_true.flatten(),
                y=Y_predicted.flatten(),
                z=Z.flatten(),
                mode="markers",
            )
        ]
    )
    figure.update_layout(
        title=f"{loss.__class__.__name__} Loss Function",
        scene=dict(xaxis_title="y_true", yaxis_title="y_predicted", zaxis_title="Loss"),
    )
    figure.show()


def _visualize_classification(loss: LossFunction, *, points: int) -> None:
    figure = Figure()
    p_1 = np.linspace(0.01, 0.99, points)
    p_2 = 1 - p_1

    y_true_options = [np.array([1, 0]), np.array([0, 1])]

    for y_true in y_true_options:
        Z = np.zeros(points)
        for i in range(points):
            y_predicted = np.array([p_1[i], p_2[i]])
            Z[i] = loss.loss(y_true=y_true, y_predicted=y_predicted)

        figure.add_trace(
            Scatter3d(
                x=p_1,
                y=p_2,
                z=Z,
                mode="lines+markers",
                name=f"y_true = {y_true}",
                marker=dict(size=3),
            )
        )

    figure.update_layout(
        title=f"{loss.__class__.__name__} Loss Function for 2 Classes",
        scene=dict(
            xaxis_title="Probability of Class 1",
            yaxis_title="Probability of Class 2",
            zaxis_title="Loss",
        ),
        legend_title="True Class",
    )

    figure.show()



from typing import Any, TypeAlias


Configuration: TypeAlias = dict[str, Any]


def basic_animation_configuration(
    redraw: bool = False, *, duration: int = 10
) -> Configuration:
    return dict(
        type="buttons",
        buttons=[
            play_button(redraw, duration=duration),
            pause_button(redraw),
            restart_button(redraw),
        ],
    )


def play_button(redraw: bool = True, *, duration: int = 10) -> Configuration:
    return dict(
        label="Play",
        method="animate",
        args=[
            None,
            {
                "frame": {"duration": duration, "redraw": redraw},
                "fromcurrent": True,
                "transition": {"duration": 0},
            },
        ],
    )


def pause_button(redraw: bool = True) -> Configuration:
    return dict(
        label="Pause",
        method="animate",
        args=[
            [None],
            {
                "frame": {"duration": 0, "redraw": redraw},
                "mode": "immediate",
                "transition": {"duration": 0},
            },
        ],
    )


def restart_button(redraw: bool = True) -> Configuration:
    return dict(
        label="Restart",
        method="animate",
        args=[
            None,
            {
                "frame": {"duration": 0, "redraw": redraw},
                "fromcurrent": False,
                "transition": {"duration": 0},
            },
        ],
    )



from typing import Protocol

from helpers.maths import Vector


class InputScaler(Protocol):
    def transform(self, data: Vector, /) -> Vector:
        """Transforms the input data."""
        ...


class OutputScaler(Protocol):
    def inverse_transform(self, data: Vector, /) -> Vector:
        """Scales the output data back to its original form."""
        ...



from typing import Protocol, TypeVar, Generic, Self, Any, Final
from dataclasses import dataclass, KW_ONLY, field

from helpers.maths import Vector, Matrix
from helpers.tank.control import Control, StatefulControl
from helpers.tank.state import TankStates
from helpers.tank.animation import OnFrame, do_nothing
from scipy.integrate import solve_ivp

import numpy as np

GOOD_STEP_SIZE: Final = 0.001

TankT = TypeVar("TankT", bound="Tank")


@dataclass(frozen=True)
class SimulationResults(Generic[TankT]):
    tank: TankT
    _: KW_ONLY
    time: Vector
    inlet_flow_rate: Vector
    outlet_flow_rate: Vector
    liquid_height: Vector

    def animate(self, *, take: int = 1, on_frame: OnFrame = do_nothing) -> Self:
        """Animates the simulation results.

        Args:
            take: The stride to use when animating the simulation results. A stride of 1 will animate
                every point in the simulation results.

        Returns:
            The same simulation results.
        """

        mask = np.concat(
            [np.arange(0, self.points - 2, take), np.array([self.points - 1])]
        )

        self.tank.animate(
            TankStates(
                time=self.time[mask],
                inlet_flow_rate=self.inlet_flow_rate[mask],
                liquid_height=self.liquid_height[mask],
                outlet_flow_rate=self.outlet_flow_rate[mask],
            ),
            on_frame=on_frame,
        )

        return self

    @property
    def points(self) -> int:
        return len(self.time)

    @property
    def final_liquid_height(self) -> float:
        return self.liquid_height[-1]


@dataclass(frozen=True)
class IntermediateValueRecorder:
    t: list[float] = field(default_factory=list)
    inlet_flow_rate: list[float] = field(default_factory=list)
    outlet_flow_rate: list[float] = field(default_factory=list)

    def record(
        self, *, t: float, inlet_flow_rate: float, outlet_flow_rate: float
    ) -> None:
        self.t.append(t)
        self.inlet_flow_rate.append(inlet_flow_rate)
        self.outlet_flow_rate.append(outlet_flow_rate)

    def inlet_flow_rates_for(self, t: Vector) -> Vector:
        return np.interp(t, self.t, self.inlet_flow_rate)

    def outlet_flow_rates_for(self, t: Vector) -> Vector:
        return np.interp(t, self.t, self.outlet_flow_rate)


class TankDynamicsJacobian(Protocol):
    def __call__(self, t: float, state: Vector) -> Matrix:
        """The Jacobian matrix of the right-hand side of the tank dynamics.

        Args:
            t: The current time point.
            state: The current state of the robot arm.

        Returns:
            The Jacobian matrix of the right-hand side of the tank dynamics. This matrix must be
            $n \\times n$, where $n$ is the number of states in the system. Entry $(i, j)$ of the
            matrix must be the partial derivative of the $i$-th element of the right-hand side with
            respect to the $j$-th state variable.
        """
        ...


class TankDynamics(Protocol):
    def __call__(self, t: float, state: Vector) -> Vector:
        """The dynamics of the tank at the given time point.

        Args:
            t: The current time point.
            state: The current state of the tank.

        Returns:
            The derivative of the state vector.

        Note:
            The state vector is typically just the liquid height in the tank.
        """
        ...

    def jacobian(self) -> TankDynamicsJacobian | None:
        """Returns a function that computes the Jacobian of the right-hand side of the tank dynamics.

        Returns:
            A function that computes the Jacobian of the right-hand side of the tank dynamics or None. If
            None, the Jacobian will be approximated numerically.
        """
        ...


class Tank(Protocol):
    def animate(self, states: TankStates, *, on_frame: OnFrame) -> Any:
        """Animates the tank over the specified time points."""
        ...

    def dynamics_for(
        self, control: Control, recorder: IntermediateValueRecorder
    ) -> TankDynamics:
        """Returns the dynamics of the tank for the given control signal."""
        ...

    def initial_conditions(self) -> Vector:
        """Returns the initial state of the tank as a state vector."""
        ...

    def results_for(
        self,
        t_range: tuple[float, float],
        recorded_points: int,
        t: Vector,
        inflow_rate: Vector,
        outflow_rate: Vector,
        solution: Matrix,
    ) -> "SimulationResults[Self]":
        """Returns the simulation results for the tank."""
        ...


class Solver(Protocol):
    def solve(
        self,
        dynamics: TankDynamics,
        initial_conditions: Vector,
        *,
        t_range: tuple[float, float],
        t: Vector,
    ) -> Matrix:
        """Solves the initial value problem for the given tank dynamics."""
        ...


class ScipySolver:
    def solve(
        self,
        dynamics: TankDynamics,
        initial_conditions: Vector,
        *,
        t_range: tuple[float, float],
        t: Vector,
    ) -> Matrix:
        return solve_ivp(
            fun=dynamics,
            jac=dynamics.jacobian(),
            y0=initial_conditions,
            t_span=t_range,
            t_eval=t,
            method="LSODA",
        ).y


class EulerSolver:
    def solve(
        self,
        dynamics: TankDynamics,
        initial_conditions: Vector,
        *,
        t_range: tuple[float, float],
        t: Vector,
    ) -> Matrix:
        y = initial_conditions
        delta_t = t[1] - t[0]

        solution = np.zeros((len(y), len(t)))
        solution[:, 0] = y

        for i in range(1, len(t)):
            y = y + delta_t * dynamics(t[i], y)
            solution[:, i] = y

        return solution


class SimulateTankMixin:
    def simulate(
        self: TankT,
        control: Control | StatefulControl,
        t_range: tuple[float, float],
        *,
        recorded_points: int | None = None,
        solver: Solver = EulerSolver(),
    ) -> SimulationResults[TankT]:
        """Simulate the tank with the given control signal.

        Args:
            control: The control signal to apply to the tank.
            t_range: The range of time points to simulate over.
            recorded_points: The number of points to use for simulating the tank. If None, the number
                of points will be determined automatically to match a step size of 1 ms.
            solver: The solver to use for solving the initial value problem. By default, the Euler
                method is used, since the dynamics are typically not stiff.

        Returns:
            The simulation results for the tank.
        """
        assert recorded_points is None or recorded_points > 1, (
            f"The number of recorded points must be greater than 1. Got: {recorded_points}"
        )

        assert t_range[0] < t_range[1], (
            f"The time range is incorrectly ordered. Got: {t_range}"
        )

        if isinstance(control, StatefulControl):
            control.reset()

        time_span = t_range[1] - t_range[0]
        recorded_points = recorded_points or int(time_span / GOOD_STEP_SIZE)
        t = np.linspace(*t_range, recorded_points)

        recorder = IntermediateValueRecorder()
        dynamics = self.dynamics_for(control, recorder)

        solution = solver.solve(
            dynamics=dynamics,
            initial_conditions=self.initial_conditions(),
            t_range=t_range,
            t=t,
        )

        return self.results_for(
            t_range=t_range,
            recorded_points=recorded_points,
            t=t,
            inflow_rate=recorder.inlet_flow_rates_for(t),
            outflow_rate=recorder.outlet_flow_rates_for(t),
            solution=solution,
        )



from dataclasses import dataclass, KW_ONLY
from helpers.network import DataMapper

from numpy import gradient
from pandas import DataFrame
from plotly.subplots import make_subplots
from plotly.graph_objects import Scatter


class infer:
    @staticmethod
    def joint_accelerations(
        *, time: str, accelerations_from: dict[str, str]
    ) -> DataMapper:
        def mapping(data: DataFrame) -> DataFrame:
            return data.assign(
                **{
                    acceleration: gradient(data[velocity], data[time])
                    for acceleration, velocity in accelerations_from.items()
                }
            )

        return mapping


@dataclass(frozen=True)
class MeasurementColumns:
    _: KW_ONLY
    angle: str
    velocity: str
    acceleration: str


def visualize(
    dynamics_data: DataFrame,
    *,
    time: str,
    measurements: list[MeasurementColumns],
    title: str,
) -> None:
    time_data = dynamics_data[time]

    figure = make_subplots(
        rows=len(measurements),
        cols=3,
        shared_xaxes=True,
        subplot_titles=[
            subplot
            for measurement in measurements
            for subplot in (
                measurement.angle,
                measurement.velocity,
                measurement.acceleration,
            )
        ],
        vertical_spacing=0.1,
    )

    for joint, measurement in enumerate(measurements, start=1):
        figure.add_trace(
            Scatter(
                x=time_data,
                y=dynamics_data[measurement.angle],
                mode="lines",
                name=measurement.angle,
                legendgroup=f"Joint {joint}",
            ),
            row=joint,
            col=1,
        )
        figure.add_trace(
            Scatter(
                x=time_data,
                y=dynamics_data[measurement.velocity],
                mode="lines",
                name=measurement.velocity,
                legendgroup=f"Joint {joint}",
            ),
            row=joint,
            col=2,
        )
        figure.add_trace(
            Scatter(
                x=time_data,
                y=dynamics_data[measurement.acceleration],
                mode="lines",
                name=measurement.acceleration,
                legendgroup=f"Joint {joint}",
            ),
            row=joint,
            col=3,
        )

    figure.update_layout(title_text=title, legend_tracegroupgap=90)
    figure.update_xaxes(title_text="Time (s)", row=len(measurements), col=1)

    figure.show()



from typing import Sequence, Final
from dataclasses import dataclass, KW_ONLY, field

from helpers.maths import Vector
from helpers.robot.simulate.results import SimulationResults, RobotT
from helpers.robot.simulate.solver import LsodaSolver
from helpers.robot.simulate.types import (
    ControlSignal,
    StatefulControlSignal,
    State,
    Position,
    Jacobian,
    Gravity,
)

from scipy.interpolate import interp1d

import numpy as np


GOOD_STEP_SIZE: Final[float] = 0.001


@dataclass
class ControlSignalRecorder:
    _: KW_ONLY
    target_t: float
    delta_t: float

    tau: Final[ControlSignal]
    tolerance: Final[float]

    recorded_t: list[float] = field(default_factory=list)
    recorded_tau: list[Vector] = field(default_factory=list)

    @staticmethod
    def recording(
        tau: ControlSignal,
        *,
        t: Vector,
        tolerance: float = 1e-10,
    ) -> "ControlSignalRecorder":
        delta_t = t[1] - t[0]

        return ControlSignalRecorder(
            target_t=t[0],
            delta_t=delta_t,
            tau=tau,
            tolerance=tolerance,
        )

    def __call__(
        self, *, q: State, p: Position, J: Jacobian, g: Gravity, t: float
    ) -> Vector:
        tau = self.tau(q=q, p=p, J=J, g=g, t=t)

        if self.target_t - self.tolerance < t:
            self._record(t, tau)

        return tau

    def torques(self, t: Vector, *, skip_first: bool) -> list[Vector]:
        # We may have not gotten every time point, due to how the ODE solver works, so
        # we will now linearly interpolate the missing values. Each element of the recorded_tau
        # list corresponds to full control signal for a given time point.
        if skip_first:
            recorded_t = self.recorded_t[1:]
            recorded_tau = self.recorded_tau[1:]
        else:
            recorded_t = self.recorded_t
            recorded_tau = self.recorded_tau

        recorded_t = np.array(recorded_t)
        recorded_tau = np.array(recorded_tau)

        # TODO: interp1d is deprecated, so it should be replaced with something else.
        interpolator = interp1d(
            recorded_t,
            recorded_tau,
            bounds_error=False,
            fill_value=(recorded_tau[0], recorded_tau[-1]),  # type: ignore
            axis=0,
            kind="linear" if len(recorded_t) > 1 else "zero",
            assume_sorted=True,
            copy=False,
        )

        return [row for row in interpolator(t).T]

    def _record(self, t: float, tau: Vector) -> None:
        self.recorded_t.append(t)
        self.recorded_tau.append(tau)
        self.target_t += self.delta_t


class RobotArmSimulationMixin:
    def simulate(
        self: RobotT,
        tau: ControlSignal | StatefulControlSignal,
        *,
        t_range: tuple[float, float],
        recorded_points: int | None = None,
        initial_velocities: Sequence[float] | None = None,
        initial_conditions: Sequence[float] | Vector | None = None,
        skip_first: bool = False,
    ) -> SimulationResults[RobotT]:
        """Simulates the robot arm dynamics.

        Args:
            tau: The input torques to the robot arm.
            t_range: The time range to simulate the robot arm dynamics over.
            recorded_points: The number of points to record the simulation results at (determined automatically by default).
            initial_velocities: The initial velocities of the robot arm joints (defaults to all zeros).
            initial_conditions: The initial conditions of the robot arm as a state vector (optional).
            skip_first: Whether to skip the first recorded point (default is `False`). This is useful
                when combining multiple simulations together, but it's mostly used internally.

        Returns:
            The simulation results for the robot arm dynamics.

        Note:
            If the initial conditions are not provided, the ones provided by the robot arm will be used. Also,
            if you are specifying the initial conditions, specifying the initial velocities is no longer necessary
            or possible. These should be included in the initial conditions vector.
        """
        assert recorded_points is None or recorded_points > 1, (
            f"The number of recorded points must be greater than 1. Got: {recorded_points}"
        )

        assert t_range[0] < t_range[1], (
            f"The time range is incorrectly ordered. Got: {t_range}"
        )

        if isinstance(tau, StatefulControlSignal):
            tau.reset()

        time_span = t_range[1] - t_range[0]
        recorded_points = recorded_points or int(time_span / GOOD_STEP_SIZE)
        t = np.linspace(*t_range, recorded_points)

        recorder = ControlSignalRecorder.recording(tau, t=t)
        dynamics = self.dynamics_for(recorder)
        solver = dynamics.solver() or LsodaSolver()

        solution = solver(
            dynamics=dynamics,
            initial_conditions=(
                initial_conditions
                if initial_conditions is not None
                else self.initial_conditions(initial_velocities=initial_velocities)
            ),
            t_range=t_range,
            t_evaluation=t,
        )

        if skip_first:
            t = t[1:]
            solution = solution[:, 1:]
            recorded_points -= 1

        return self.results_from(
            t_range=t_range,
            recorded_points=recorded_points,
            t=t,
            tau=recorder.torques(t, skip_first=skip_first),
            solution=solution,
        )



from typing import Protocol, Sequence, TypeVar, Generic, Callable, cast
from dataclasses import dataclass, field, KW_ONLY
from functools import cached_property
from helpers.maths import Matrix, Vector, Point
from helpers.robot.link import Link
from helpers.robot.joint import Joint
from helpers.robot.simulate import (
    SimulatableRobotArm,
    SimulationResults,
    ControlSignal,
    State,
    PrecomputedState,
    LazyPosition,
    LazyJacobian,
    PrecomputedGravity,
)

import numpy as np
from scipy.optimize import approx_fprime

PlanarRobotArmT = TypeVar("PlanarRobotArmT", bound="PlanarRobotArm")
SimulatableRobotT = TypeVar("SimulatableRobotT", bound="SimulatableRobotArm")


class PlanarRobotArm(SimulatableRobotArm, Protocol):
    def forward_kinematics(self, q: Sequence[float] | Vector) -> list[Point]:
        """Calculates the forward kinematics of the robot arm for the given joint angles."""
        ...

    def jacobian(self, q: Sequence[float] | Vector) -> Matrix:
        """Returns the Jacobian matrix of the robot arm at the given configuration."""
        ...

    @property
    def link_1(self) -> Link:
        """The first link of the robot arm."""
        ...

    @property
    def link_2(self) -> Link:
        """The second link of the robot arm."""
        ...

    @property
    def joint_1(self) -> Joint:
        """The first joint of the robot arm."""
        ...

    @property
    def joint_2(self) -> Joint:
        """The second joint of the robot arm."""
        ...


@dataclass(frozen=True)
class DynamicsContext:
    """This class contains all precomputed values that are needed to calculate the dynamics of the
    planar robot arm.

    Attributes:
        c_1: $cos(q_1)$
        c_1_2: $cos(q_1 + q_2)$
        c_2: $cos(q_2)$
        s_2: $sin(q_2)$
        q_dot_1: $\\dot{q}_1$
        q_dot_2: $\\dot{q}_2$
        m_1: The mass of the first link.
        m_2: The mass of the second link.
        l_1: The length of the first link.
        l_2: The length of the second link.
        l_c1: The distance from the base of the first link to its center of mass.
        l_c2: The distance from the base of the second link to its center of mass.
        mu_1: The friction of the first joint.
        mu_2: The friction of the second joint.
        I_c1: The moment of inertia of the first link.
        I_c2: The moment of inertia of the second link.
        g: The acceleration due to gravity.

    Note:
        $q_1$ and $q_2$ are the first and second joint angles of the robot arm, respectively.
    """

    _: KW_ONLY
    c_1: float
    c_1_2: float
    c_2: float
    s_2: float
    q_dot_1: float
    q_dot_2: float
    m_1: float
    m_2: float
    l_1: float
    l_2: float
    l_c1: float
    l_c2: float
    mu_1: float
    mu_2: float
    I_c1: float
    I_c2: float
    g: float


@dataclass(frozen=True)
class DynamicsJacobianContext:
    """This class contains all precomputed values that are needed to calculate the Jacobian matrix of
    the right-hand side of the dynamics equation of the planar robot arm.

    Attributes:
        alpha: $\\alpha = m_2 * l_1 * l_c2$
        beta: $\\beta = m_2 * g * l_c2$
        c_2: $cos(q_2)$
        s_1: $sin(q_1)$
        s_1_2: $sin(q_1 + q_2)$
        s_2: $sin(q_2)$
        q_dot_1: $\\dot{q}_1$
        q_dot_2: $\\dot{q}_2$
        m_1: The mass of the first link.
        m_2: The mass of the second link.
        l_1: The length of the first link.
        l_c1: The distance from the base of the first link to its center of mass.
        l_c2: The distance from the base of the second link to its center of mass.
        mu_1: The friction of the first joint.
        mu_2: The friction of the second joint.
        g: The acceleration due to gravity.

    Note:
        $q_1$ and $q_2$ are the first and second joint angles of the robot arm, respectively.
    """

    _: KW_ONLY
    alpha: float
    beta: float
    c_2: float
    s_1: float
    s_1_2: float
    s_2: float
    q_dot_1: float
    q_dot_2: float
    m_1: float
    m_2: float
    l_1: float
    l_c1: float
    l_c2: float
    mu_1: float
    mu_2: float
    g: float


@dataclass(frozen=True)
class DynamicsContextProvider:
    _: KW_ONLY
    robot: PlanarRobotArm
    I_c1: float
    I_c2: float
    g: float = 9.81

    @staticmethod
    def create_for(robot: PlanarRobotArm) -> "DynamicsContextProvider":
        m_1, delta_1 = robot.link_1.mass, robot.link_1.com_size
        m_2, delta_2 = robot.link_2.mass, robot.link_2.com_size

        return DynamicsContextProvider(
            robot=robot,
            I_c1=m_1 * delta_1**2,
            I_c2=m_2 * delta_2**2,
        )

    def __call__(self, q: Vector, q_dot: Vector) -> DynamicsContext:
        q_1, q_2 = q
        q_dot_1, q_dot_2 = q_dot

        return DynamicsContext(
            c_1=np.cos(q_1),
            c_1_2=np.cos(q_1 + q_2),
            c_2=np.cos(q_2),
            s_2=np.sin(q_2),
            q_dot_1=q_dot_1,
            q_dot_2=q_dot_2,
            m_1=self.robot.link_1.mass,
            m_2=self.robot.link_2.mass,
            l_1=self.robot.link_1.length,
            l_2=self.robot.link_2.length,
            l_c1=self.robot.link_1.distance_to_com,
            l_c2=self.robot.link_2.distance_to_com,
            mu_1=self.robot.joint_1.friction,
            mu_2=self.robot.joint_2.friction,
            I_c1=self.I_c1,
            I_c2=self.I_c2,
            g=self.g,
        )

    def jacobian(self, q: Vector, q_dot: Vector) -> DynamicsJacobianContext:
        q_1, q_2 = q
        q_dot_1, q_dot_2 = q_dot
        m_2 = self.robot.link_2.mass
        l_1, l_c2 = self.robot.link_1.length, self.robot.link_2.distance_to_com

        return DynamicsJacobianContext(
            alpha=m_2 * l_1 * l_c2,
            beta=m_2 * self.g * l_c2,
            c_2=np.cos(q_2),
            s_1=np.sin(q_1),
            s_1_2=np.sin(q_1 + q_2),
            s_2=np.sin(q_2),
            q_dot_1=q_dot_1,
            q_dot_2=q_dot_2,
            m_1=self.robot.link_1.mass,
            m_2=m_2,
            l_1=l_1,
            l_c1=self.robot.link_1.distance_to_com,
            l_c2=l_c2,
            mu_1=self.robot.joint_1.friction,
            mu_2=self.robot.joint_2.friction,
            g=self.g,
        )


@dataclass
class LazyGravity:
    context: DynamicsContextProvider
    gravity_provider: Callable[[DynamicsContext], Vector]
    t: float
    g: Vector | None = field(default=None, init=False)

    def __call__(self, q: State) -> Vector:
        if self.g is None:
            self.g = self.gravity_provider(self.context(q(self.t), q.dot(self.t)))

        return self.g


# TODO: Document how the Jacobian is calculated.
@dataclass(frozen=True)
class RigidBodyDynamicsJacobian(Generic[PlanarRobotArmT]):
    dynamics: "RigidBodyDynamics[PlanarRobotArmT]"

    def __call__(self, t: float, state: Vector) -> Matrix:
        q, q_dot = state[:2], state[2:]
        J = self.jacobian_matrix(q, q_dot, t)
        return J

    def jacobian_matrix(self, q: Vector, q_dot: Vector, t: float) -> Matrix:
        context = self.dynamics.context(q, q_dot)
        D_inv = self.dynamics.D_inv(context)
        C = self.dynamics.C(context)
        g = self.dynamics.g(context)
        mu = self.dynamics.mu(context)

        jacobian_context = self.dynamics.context.jacobian(q, q_dot)
        J_D_inv = self.J_D_inv(jacobian_context)
        J_C = self.J_C(jacobian_context)
        J_g = self.J_g(jacobian_context)
        J_tau = self.J_tau(q, q_dot, t)
        J_tau_mu = self.J_tau_mu(jacobian_context)
        J_q_dot = self.J_q_dot

        tau = self.dynamics.tau(
            q=PrecomputedState(q=q, q_dot=q_dot),
            p=LazyPosition(robot=self.dynamics.robot, t=t),
            J=LazyJacobian(robot=self.dynamics.robot, t=t),
            g=PrecomputedGravity(g=g),
            t=t,
        )
        tau_mu = -mu @ q_dot

        return np.vstack(
            [
                np.array(
                    [
                        [0, 0, 1, 0],
                        [0, 0, 0, 1],
                    ]
                ),
                np.vstack(
                    [
                        -D_inv @ D_inv_x @ D_inv @ (tau + tau_mu - C @ q_dot - g)
                        + (D_inv @ (tau_x + tau_mu_x - C_x @ q_dot - C @ q_dot_x - g_x))
                        for D_inv_x, C_x, g_x, tau_x, tau_mu_x, q_dot_x in zip(
                            J_D_inv, J_C, J_g, J_tau, J_tau_mu, J_q_dot
                        )
                    ]
                ).T,
            ]
        )

    def J_tau(self, q: Vector, q_dot: Vector, t: float) -> Sequence[Vector]:
        return cast(
            Sequence[Vector],
            approx_fprime(np.concat([q, q_dot]), self.tau_for, 1e-7, t).T.tolist(),
        )

    def J_tau_mu(self, context: DynamicsJacobianContext) -> Sequence[Vector]:
        mu_1, mu_2 = context.mu_1, context.mu_2

        J_tau_mu_1 = J_tau_mu_2 = np.zeros(2)
        J_tau_mu_3 = np.array([-mu_1, 0])
        J_tau_mu_4 = np.array([0, -mu_2])

        return [J_tau_mu_1, J_tau_mu_2, J_tau_mu_3, J_tau_mu_4]

    def J_D_inv(self, context: DynamicsJacobianContext) -> Sequence[Matrix]:
        alpha = context.alpha
        s_2 = context.s_2

        D_x_1 = D_x_3 = D_x_4 = np.zeros((2, 2))
        D_x_2 = alpha * (-s_2) * np.array([[2, 1], [1, 0]])

        return [D_x_1, D_x_2, D_x_3, D_x_4]

    def J_C(self, context: DynamicsJacobianContext) -> Sequence[Matrix]:
        alpha = context.alpha
        c_2 = context.c_2
        s_2 = context.s_2
        q_dot_1, q_dot_2 = context.q_dot_1, context.q_dot_2

        C_x_1 = np.zeros((2, 2))
        C_x_2 = alpha * c_2 * np.array([[-q_dot_2, -(q_dot_1 + q_dot_2)], [q_dot_1, 0]])
        C_x_3 = alpha * s_2 * np.array([[0, -1], [1, 0]])
        C_x_4 = alpha * s_2 * np.array([[-1, -1], [0, 0]])

        return [C_x_1, C_x_2, C_x_3, C_x_4]

    def J_g(self, context: DynamicsJacobianContext) -> Sequence[Vector]:
        beta = context.beta
        s_1, s_1_2 = context.s_1, context.s_1_2
        m_1, m_2 = context.m_1, context.m_2
        l_1 = context.l_1
        l_c1, l_c2 = context.l_c1, context.l_c2
        g = context.g

        g_x_1 = -g * np.array(
            [m_1 * l_c1 * s_1 + m_2 * (l_1 * s_1 + l_c2 * s_1_2), m_2 * l_c2 * s_1_2]
        )
        g_x_2 = -beta * s_1_2 * np.array([1, 1])
        g_x_3 = g_x_4 = np.zeros(2)

        return [g_x_1, g_x_2, g_x_3, g_x_4]

    @cached_property
    def J_q_dot(self) -> Sequence[Vector]:
        q_dot_x_1 = q_dot_x_2 = np.zeros(2)
        q_dot_x_3 = np.array([1, 0])
        q_dot_x_4 = np.array([0, 1])

        return [q_dot_x_1, q_dot_x_2, q_dot_x_3, q_dot_x_4]

    def tau_for(self, x: Vector, t: float, /) -> Vector:
        q, q_dot = x[:2], x[2:]

        return self.dynamics.tau(
            q=PrecomputedState(q=q, q_dot=q_dot),
            p=LazyPosition(robot=self.dynamics.robot, t=t),
            J=LazyJacobian(robot=self.dynamics.robot, t=t),
            g=LazyGravity(
                context=self.dynamics.context, gravity_provider=self.dynamics.g, t=t
            ),
            t=t,
        )


# TODO: Document how the dynamics are calculated.
@dataclass(frozen=True)
class RigidBodyDynamics(Generic[PlanarRobotArmT]):
    tau: ControlSignal
    robot: PlanarRobotArmT
    context: DynamicsContextProvider

    @staticmethod
    def create_for(
        robot: PlanarRobotArmT, *, torque: ControlSignal
    ) -> "RigidBodyDynamics[PlanarRobotArmT]":
        return RigidBodyDynamics(
            tau=torque,
            robot=robot,
            context=DynamicsContextProvider.create_for(robot),
        )

    @staticmethod
    def results_for(
        robot: SimulatableRobotT,
        *,
        t_range: tuple[float, float],
        recorded_points: int,
        t: Vector,
        tau: list[Vector],
        solution: Matrix,
    ) -> SimulationResults[SimulatableRobotT]:
        q = [row % (2 * np.pi) for row in solution[:2]]
        q_dot = [row for row in solution[2:]]
        return SimulationResults.of(
            robot,
            t_range=t_range,
            recorded_points=recorded_points,
            t=t,
            tau=tau,
            q=q,
            q_dot=q_dot,
            final_state=solution[:, -1],
        )

    def __call__(self, t: float, state: Vector) -> Vector:
        q, q_dot = state[:2], state[2:]
        q_dot_dot = self.dynamics(q, q_dot, t)
        return np.concatenate([q_dot, q_dot_dot])

    def dynamics(self, q: Vector, q_dot: Vector, t: float) -> Vector:
        context = self.context(q, q_dot)
        D_inv, C, g, mu = (
            self.D_inv(context),
            self.C(context),
            self.g(context),
            self.mu(context),
        )

        tau = self.tau(
            q=PrecomputedState(q=q, q_dot=q_dot),
            p=LazyPosition(robot=self.robot, t=t),
            J=LazyJacobian(robot=self.robot, t=t),
            g=PrecomputedGravity(g=g),
            t=t,
        )
        tau_mu = -mu @ q_dot

        q_dot_dot = D_inv @ (tau + tau_mu - C @ q_dot - g).flatten()
        return q_dot_dot

    def jacobian(self) -> "RigidBodyDynamicsJacobian[PlanarRobotArmT]":
        return RigidBodyDynamicsJacobian(self)

    def solver(self) -> None:
        # The default solver is just fine.
        return None

    def D_inv(self, context: DynamicsContext) -> Matrix:
        c_2 = context.c_2
        m_1, m_2 = context.m_1, context.m_2
        l_1 = context.l_1
        l_c1, l_c2 = context.l_c1, context.l_c2
        I_c1, I_c2 = context.I_c1, context.I_c2

        d_11 = (
            m_1 * l_c1**2
            + m_2 * (l_1**2 + l_c2**2 + 2 * l_1 * l_c2 * c_2)
            + I_c1
            + I_c2
        )
        d_12 = m_2 * (l_c2**2 + l_1 * l_c2 * c_2) + I_c2
        d_22 = m_2 * l_c2**2 + I_c2
        det_D = d_11 * d_22 - d_12**2

        return (
            np.array(
                [
                    [d_22, -d_12],
                    [-d_12, d_11],
                ]
            )
            / det_D
        )

    def C(self, context: DynamicsContext) -> Matrix:
        q_dot_1, q_dot_2 = context.q_dot_1, context.q_dot_2
        s_2 = context.s_2
        m_2 = context.m_2
        l_1 = context.l_1
        l_c2 = context.l_c2

        return np.array(
            [
                [
                    -m_2 * l_1 * l_c2 * s_2 * q_dot_2,
                    (-m_2 * l_1 * l_c2 * s_2 * (q_dot_1 + q_dot_2)),
                ],
                [m_2 * l_1 * l_c2 * s_2 * q_dot_1, 0],
            ]
        )

    def g(self, context: DynamicsContext) -> Vector:
        c_1 = context.c_1
        c_1_2 = context.c_1_2
        m_1, m_2 = context.m_1, context.m_2
        l_1 = context.l_1
        l_c1, l_c2 = context.l_c1, context.l_c2
        g = context.g

        return np.array(
            [
                m_1 * g * l_c1 * c_1 + m_2 * g * (l_1 * c_1 + l_c2 * c_1_2),
                m_2 * g * l_c2 * c_1_2,
            ]
        )

    def mu(self, context: DynamicsContext) -> Matrix:
        mu_1, mu_2 = context.mu_1, context.mu_2

        return np.array([[mu_1, 0], [0, mu_2]])



from typing import Protocol
from dataclasses import dataclass, field

from helpers.rl import LearningAgent
from helpers.tank.discretize import TankSystemState, TankSystemAction


@dataclass(frozen=True)
class TrainingResult:
    learning_rate: float
    starting_height: float
    setpoint: float
    height_difference: float
    accumulated_reward: float


class RewardFunction(Protocol):
    def __call__(self, *, state: TankSystemState) -> float:
        """Returns the reward for transitioning to the given state."""
        ...


class SetpointProvider(Protocol):
    def __call__(self, *, at: float) -> float:
        """Returns the setpoint for the given time."""
        ...


@dataclass(frozen=True)
class SetpointRecorder:
    setpoint: SetpointProvider
    times: list[float] = field(default_factory=list, init=False)
    setpoints: list[float] = field(default_factory=list, init=False)

    def __call__(self, *, at: float) -> float:
        setpoint = self.setpoint(at=at)
        self.times.append(at)
        self.setpoints.append(setpoint)

        return setpoint


@dataclass
class LearningController:
    agent: LearningAgent[TankSystemState, TankSystemAction]
    reward: RewardFunction
    setpoint: SetpointProvider

    accumulated_reward: float = field(default=0, init=False)
    last_time_step: float | None = field(default=None, init=False)
    last_state: TankSystemState | None = field(default=None, init=False)
    last_action: TankSystemAction | None = field(default=None, init=False)

    def __call__(self, *, t: float, h: float) -> float:
        self.verify(t)

        setpoint = self.setpoint(at=t)
        state = TankSystemState(height_difference=setpoint - h)
        self.update(state)

        self.last_state = state
        self.last_action = self.agent(state)

        return self.last_action.inflow_rate

    def verify(self, t: float) -> None:
        assert self.last_time_step is None or t > self.last_time_step, (
            f"The time must be strictly increasing, but {t} is not greater than "
            f"the last time step {self.last_time_step}."
        )

    def update(self, state: TankSystemState) -> None:
        if self.last_state is None or self.last_action is None:
            return

        reward = self.reward(state=state)

        self.accumulated_reward += reward
        self.agent.update(
            state=self.last_state,
            action=self.last_action,
            next_state=state,
            reward=reward,
        )

    def reset(self) -> None:
        self.accumulated_reward = 0
        self.last_time_step = None
        self.last_state = None
        self.last_action = None



from typing import TypeAlias, NamedTuple
from numpy.typing import NDArray
import numpy as np

Matrix: TypeAlias = NDArray[np.floating]
Vector: TypeAlias = NDArray[np.floating]
IntMatrix: TypeAlias = NDArray[np.integer]
IntVector: TypeAlias = NDArray[np.integer]
BoolMatrix: TypeAlias = NDArray[np.bool_]


class Point(NamedTuple):
    """A point in two dimensions.

    Attributes:
        x: The x-coordinate of the point.
        y: The y-coordinate of the point.

    Example:
        You can make a point (1, 2) in 2D space, where the x coordinate is 1 and the y coordinate is 2, like this:

        ```python
        point = Point(1, 2)
        print(point)
        # Output: Point(x=1, y=2)
        ```
    """

    x: float
    y: float

    def vector(self) -> Vector:
        """Returns the point as a vector."""
        return np.array([self.x, self.y])

    def is_approximately(self, other: "Point", *, tolerance: float = 1e-10) -> bool:
        """Checks if the point is approximately equal to another point.

        Args:
            other: The other point to compare with.
            tolerance: The relative tolerance for the comparison (default is 1e-10).

        Returns:
            True if the points are approximately equal, False otherwise.
        """
        return np.allclose(self.vector(), other.vector(), rtol=tolerance)



from typing import Protocol, Sequence
from dataclasses import dataclass, field, KW_ONLY

from helpers.maths import Matrix, Vector, Point
from helpers.robot.simulate.simulate import (
    State,
)


class RobotKinematics(Protocol):
    def forward_kinematics(self, q: Vector) -> Sequence[Point]:
        """Computes the forward kinematics of the robot for the given configuration."""
        ...

    def jacobian(self, q: Vector) -> Matrix:
        """Computes the Jacobian of the robot for the given configuration."""
        ...


@dataclass(frozen=True)
class PrecomputedState:
    _: KW_ONLY
    q: Vector
    q_dot: Vector

    def __call__(self, t: float) -> Vector:
        return self.q

    def dot(self, t: float) -> Vector:
        return self.q_dot


@dataclass
class LazyPosition:
    robot: RobotKinematics
    t: float
    p: Vector | None = field(default=None, init=False)

    def __call__(self, q: State) -> Vector:
        if self.p is None:
            self.p = self.robot.forward_kinematics(q(self.t))[-1].vector()

        return self.p


@dataclass
class LazyJacobian:
    robot: RobotKinematics
    t: float
    J: Matrix | None = field(default=None, init=False)

    def __call__(self, q: State) -> Matrix:
        if self.J is None:
            self.J = self.robot.jacobian(q(self.t))

        return self.J


@dataclass(frozen=True)
class PrecomputedGravity:
    _: KW_ONLY
    g: Vector

    def __call__(self, q: State) -> Vector:
        return self.g


class UnknownGravity:
    def __call__(self, q: State) -> Vector:
        raise ValueError("Gravity vector is unknown.")



from . import control as control, simulate as simulate
from .arm import (
    RobotArmMixin as RobotArmMixin,
    JointPositionsFunction as JointPositionsFunction,
)
from .link import Link as Link
from .joint import Joint as Joint
from .animation import (
    JointAngles as JointAngles,
    RobotAnimator as RobotAnimator,
)
from .inverse import trajectory_joint_angles_for as trajectory_joint_angles_for
from .system import system_matrices_for as system_matrices_for
from .planar import PlanarRobotArm as PlanarRobotArm
from .simulate import (
    ControlSignal as ControlSignal,
    RobotArmSimulationMixin as RobotArmSimulationMixin,
    SimulationResults as SimulationResults,
    DynamicsSolver as DynamicsSolver,
    RobotDynamics as RobotDynamics,
    SimulatableRobotArm as SimulatableRobotArm,
    infer as infer,
    visualize as visualize,
    MeasurementColumns as MeasurementColumns,
)
from .jacobian import RobotJacobianMixin as RobotJacobianMixin
from .control import (
    ControlSignalProvider as ControlSignalProvider,
    Trajectory as Trajectory,
    trajectories as trajectories,
)



from .app import GridGameApp as GridGameApp
from .game import (
    GridGame as GridGame,
    Position as Position,
    Move as Move,
    Moves as Moves,
    RewardFunction as RewardFunction,
    TransitionFunction as TransitionFunction,
    EndGame as EndGame,
)
from .chance import try_with as try_with
from .table import (
    QTable as QTable,
    SimpleQTable as SimpleQTable,
    ZeroInitializer as ZeroInitializer,
    ConstantInitializer as ConstantInitializer,
    RandomInitializer as RandomInitializer,
)
from .learn import LearningAgent as LearningAgent



from typing import Protocol, runtime_checkable


class Control(Protocol):
    """This represents a control system for the tank system."""

    def __call__(self, *, t: float, h: float) -> float:
        """Returns the control signal defining the inflow rate at the given time.

        Args:
            t: The current time.
            h: The current height of the liquid in the tank.

        Returns:
            The control signal defining the inflow rate.
        """
        ...


@runtime_checkable
class StatefulControl(Control, Protocol):
    def reset(self) -> None:
        """Resets the state of the control system."""
        ...
